###################################################################
# NYISO DCR Reset 2020: Net EAS Revenues Model
# Model Control
#_________________________________________________________________;

#### Script Purpose:
####    This script is the model control script for the Net EAS revenues fossil model. Running this script executes the model by calling upon functions stored in other scripts.    

#### General Note:
####    All code in this script should be run sequentially. The code statements in this script tend to use objects defined in prior statements as inputs, 
####    so they will not run properly if those objects have not been defined already by executing the prior statements.

# _______________________________________________________

# This statement clears the R environment to allow you to execute the code cleanly
rm(list = ls())

# USER-SET TOGGLES: BASIC ------------------------------------------------
## These toggles are core toggles that users should set before running the code:
### 1.  Run folder name -- this points to the run folder subfolder in which the case specifications and any other custom case inputs are contained. 
###     For instance, if your run folder is named "NYISO_DCR", you would just put "NYISO_DCR".
runfolder_name <- ## FILL IN WITH RUNFOLDER NAME [NOT FILEPATH] ##
### 2. Custom Run ID
#### If you want to add a custom ID to the filename of each file you export with this run, set that here in quotation marks (e.g. run_ID <- "my_run"). Otherwise, set to NULL (run_ID <- NULL).
run_ID <- ## FILL IN ##

# USER-SET TOGGLES: ADVANCED ---------------------------------------------
## These toggles are additional toggles that users can set before running the code. Introductory users relying upon the readme are encouraged to leave these in their default settings.
### 3.  Time key -- set include_time to TRUE if you would like to include this in the filenames of export files. Otherwise, set to FALSE.
include_time <- FALSE
if(include_time) {
  timekey <<- format(Sys.time(), "D%Y.%m.%d T%H.%M") 
}else{
  timekey <<- NULL
}
### 4a. Intermediate datasets -- set here whether you would like to export an intermediate dataset for module 1 (TRUE) or not (FALSE)
export_mod1 <- FALSE
### 4b. Intermediate datasets -- set here whether you would like to use intermediate .rds datasets already created and exported.
#### If you do not wish to use any intermediate datasets, set both variables to FALSE
#### If you wish to use an intermediate dataset for Module 1 or for Module 2, set the corresponding toggle to TRUE and insert the filename of the intermediate set you would like
#### to use (e.g. "D2020.01.01 T10.00 Clean_Inputs.rds")
use_existing_mod1data <- FALSE
use_existing_mod2data <- FALSE
mod1data_name <- NULL
mod2data_name <- NULL

#_________________________________________________________________________________________________________________________________________________________________________

# PRELIMINARY STEPS -----------------------------------------------------

## Install and load packages ------------------------

### First, install and load devtools, which we use to install specific versions of packages
if(!("devtools" %in% installed.packages()[,"Package"])) {
  install.packages("devtools")  
}
if(!("devtools" %in% .packages())){
  library(devtools) 
}

### Then, install the other specific versions of packages as needed
packages = c("tidyverse","lubridate", "readxl", "data.table", "writexl")
package_version = c("1.3.0","1.7.4", "1.3.1", "1.12.4", "1.2")
package_list <- data.frame(packages, package_version, stringsAsFactors = FALSE)

new_packages <- package_list[which(!(package_list$packages %in% installed.packages()[,"Package"])),]
if(nrow(new_packages) > 0) map2(new_packages$packages, new_packages$package_version, function(x,y) install_version(x, version = y))

### Grab the list of all currently loaded packages 
loaded_packages = .packages()

for(i in 1:length(packages)) { 
  
  current_package = packages[i]
  
  ### If the current package is not already loaded into our session ...
  if(!(current_package %in% loaded_packages)) { 
    ### ... then load it in. The character.only option specifies that the name will be given as a string 
    library(current_package, character.only = TRUE)
  } 
}


# Load Functions, Set Paths, and Import Preliminary Data ----------------------------------------------

#### NOTE:  If operating on a Windows system, use double backslashes in place of single backslashes. R treats \ as an escape character and will not process paths correctly unless you use \\
####        to denote \. THE ONE EXCEPTION is the pattern = "\\.R" slashes on line 106; LEAVE THOSE AS IS.
####        if operating on a Linux system, use forward slashes in filepaths, but again leave line 106 as is.
path <- ## FILL IN WITH FULL FILEPATH ##
root <- paste(path, "Programs", sep = "\\")
shared_inputs_folder <- paste(path, "Shared Inputs\\", sep = "\\")
run_folder <- paste(path, "Runs", runfolder_name, sep = "\\")
run_inputs_folder <- paste(run_folder, "1. Run Inputs\\", sep = "\\")
output <- paste(run_folder, "3. Outputs\\", sep = "\\")
intermediate_datasets <- paste(run_folder, "2. Intermediate Datasets\\", sep = "\\")

## Import specifications applying to all cases (start date, intermediate dataset information) and set as variables for reference throughout
exec_specs <- read_xlsx(paste0(run_inputs_folder, "Run Specifications.xlsx"), sheet = "All_Case_Specs")
start_period <- exec_specs$Historical_Start_Period
start_date <- ymd(paste0(substr(start_period, 1, 4), "0901"))
end_date <- start_date + years(3) - days(1)
intermediate_dataset_format <- exec_specs$Intermediate_Dataset_Format
int_output_toggle <- exec_specs$Generate_Intermediate_Output

## Imports functions specified in the scripts that accompany the control file
functions = list.files(paste0(root, "\\Functions"), full.names = TRUE, pattern = "\\.R")
sapply(functions, source)

## Import general model-wide parameters and set key variables
params <- read_xlsx(paste0(shared_inputs_folder, "General Model Parameters.xlsx"), sheet = "Params_Fossil")

# __________________________________________________________________________________________________________________________________________________________;

# EXECUTE MODEL -------------------------------------------------

## 1. Implement Module 1 ------------------------------------------

### A. Define cases ----------------------------------------
all_cases <- read_xlsx(paste0(run_inputs_folder, "Run Specifications.xlsx"), sheet = "Individual_Case_Specs")
case_vec <- c(1:nrow(all_cases))
#### ADJUSTING CASE_VEC NUMBERS
# case_vec is a vector that determines which cases to evaluate in each execution of the model. It should contain the case numbers from all cases you wish to 
# evaluate as specified in workbook "Run Specifications.xlsx", sheet "Individual_Case_Specs". 
#   For instance, if you would only like to evaluate case number 2, you can set case_vec as follows:
#         case_vec <- c(2)
#   If you would like to evaluate cases 2 through 4, you can set case_vec as either of the following:
#         case_vec <- c(2,3,4)
#         case_vec <- c(2:4)
#   If you add cases on Individual_Case_Specs, you can also call them here. Say you added two more cases on Individual_Case_Specs and numbered them 6 and 7. If you wanted to 
#   evaluate cases 1, 6, and 7, you can set case_vec as ether of the following:
#         case_vec <- c(1,6,7)
#         case_vec <- c(1,6:7)
case_names_vec <- paste0("Case_", case_vec)
case_list <- lapply(case_vec, unit_import, all_cases = all_cases) %>% set_names(case_names_vec)

### B. Generate Input Base (1.1-1.3)
#### Below, we define a function (gen_input_base) that executes the functions for Modules 1.1-1.3 for a single case
gen_input_base <- function (case) {
  # Set variables to apply in function
  toggle_list <- case$Toggles
  unit_list <- case$Unit_List
  unit_type <- toggle_list$Unit_Type
  Zones <- toggle_list$Zones
  pci_name <- toggle_list$Price_Cost_Inputs_Filename
  techparams_name <- toggle_list$Technical_Inputs_Filename
  # Execute imports and merges
  price_and_cost_inputs <<- economic_inputs(unit_type, toggle_list$Price_Cost_Inputs_Filename, params)
  tech_params <<- Engineering_Import(unit_list, toggle_list$SCR_Toggle, toggle_list$Technical_Inputs_Filename)
  eas_func_base <- input_merge(price_and_cost_inputs,
                               unit_list,
                               tech_params,
                               unit_type,
                               toggle_list$Zone_C_Hub,
                               start_date)  
  case_info <- list(unit_list, case$Toggles) %>%
    set_names("Unit_List", "Toggles")
  input_sets <- list(case_info, eas_func_base) %>% set_names("Units_and_Toggles", "Inputs")
  return(input_sets)
}

#### We apply the gen_input_base function to each case specified in case_vec and export intermediate output if instructed by the user
if (!use_existing_mod1data) {
  input_base <- lapply(case_list, gen_input_base)   
  ##### The output of this is a list containing different input bases, upon which to execute the model logic.
  ##### Export an intermediate dataset at the end of Module 1 (if instructed by user)
  if(export_mod1){
    if(include_time){
      exp_name_1 <- paste(paste0(timekey, run_ID), "Clean Inputs")
    }else{
      exp_name_1 <- paste(run_ID, "Clean Inputs")
    }
    exp_datasets(int_output_toggle = int_output_toggle,
                 datasets = input_base, 
                 exp_name = exp_name_1, 
                 format = intermediate_dataset_format,
                 case_vec = case_vec)  
  }
}else{
  input_base <- readRDS(paste0(intermediate_datasets, mod1data_name))
}

## 2. Implement Module 2: unit commitment logic (2.1-2.4) -------------------------------------------
if(!use_existing_mod2data){
  ### 2A. Calculate base potential profits (2.1) ----------------------------
  base_pot_profits <- lapply(case_vec, eval_case_base, inputs = input_base) %>% set_names(case_names_vec)
  
  ### 2B. Set unit commitments (2.2-2.4) ----------------------------------------
  unit_dam_commits <- lapply(case_vec, set_dam_commits, inputs = base_pot_profits) %>% set_names(case_names_vec)

  unit_rtd_commits <- lapply(case_vec, set_rtd_commits, inputs = unit_dam_commits) %>% set_names(case_names_vec)
  
  unit_hourly_results <- lapply(case_vec, eford_plus_runtime_adj, inputs = unit_rtd_commits) %>% set_names(case_names_vec)
  
  ### Export an intermediate dataset at the end of Module 2 (if instructed by user)
  if(include_time){
    exp_name_2 <- paste(paste0(timekey, run_ID), "Model Outputs")
  }else{
    exp_name_2 <- paste(run_ID, "Model Outputs")
  }
  exp_datasets(datasets = unit_hourly_results, 
               int_output_toggle = int_output_toggle,
               exp_name = exp_name_2, 
               format = intermediate_dataset_format,
               case_vec = case_vec)
}else{
  unit_hourly_results <- readRDS(paste0(intermediate_datasets, mod2data_name))
}

## 3. Implement Module 3: generate summary tables and export to Excel (3.1) -------------------------------------------

### 3A. Generate summary tables (3.1) ---------------------------------

#### Full output diagnostic tables
diagnostic_tables <- lapply(case_vec, all_unit_summaries, inputs = unit_hourly_results) %>% set_names(case_names_vec)
result_type_vec <- names(diagnostic_tables[[1]][[1]])
stack_diagnostics <- function(diagnostic_tables, result_cat, start_date) {
  result_cat_run_diagnostics <- lapply(names(diagnostic_tables), isolate_result_type, result_cat = result_cat, start_date = start_date) %>% rbindlist(use.names = TRUE)
}
full_stacked_run_diagnostics <- lapply(result_type_vec, stack_diagnostics, diagnostic_tables = diagnostic_tables, start_date = start_date) %>% set_names(result_type_vec)

#### Isolate DCM results
dcm_results_table <- full_stacked_run_diagnostics[[result_type_vec[1]]] %>% select(-Avg.Hour.Count, -Avg.Start.Count) %>%
  mutate(Lookup = paste(Unit, Area, SCR_Toggle, Gas_Only_Toggle, sep = "_")) %>% select(Lookup, everything())

### 3B. Export Results to Excel  ----------------------------------------
#### Append readme and case specs to diagnostic outputs workbook
output_readme <- read_csv(paste0(shared_inputs_folder, "Output_Files_Readme.csv"))
full_stacked_run_diagnostics[["Output_Readme"]] <- output_readme
full_stacked_run_diagnostics[["Individual_Case_Specs"]] <- all_cases
full_stacked_run_diagnostics <- full_stacked_run_diagnostics[c("Output_Readme", "Individual_Case_Specs", result_type_vec)]
#### Export diagnostic outputs
if(is.null(run_ID) & !include_time){
  nametag <- NULL
  }else if(is.null(run_ID)){nametag <- timekey
  }else if(!include_time) {nametag <- run_ID
  }else{nametag <- paste0(timekey, run_ID)}
diag_output_filename <- paste0(output, paste(nametag, "Diagnostic Tables.xlsx"))
write_xlsx(full_stacked_run_diagnostics, diag_output_filename)
#### Export DCM input summary table
dcm_output_filename <- paste0(output, paste(nametag, "DCM Inputs.xlsx"))
write_xlsx(list(DCM_Inputs = dcm_results_table), dcm_output_filename)

