###################################################################
# NYISO DCR Reset 2020: Net EAS Revenues Model
# Module 1, part 1
#_________________________________________________________________

# Script purpose 
#   This script includes functions for importing data contained in the Price and Cost Inputs.xlsx workbook: all hourly, yearly, monthly, and daily price inputs into the model, as well as a list of dates and the LOE-AFs.
#   It also incorporates parameters around fuel transport premiums and fuel taxes into the resulting dataset
#   Inputs: a toggle for whether the case is evaluating CTs or CCs ("Unit_Type"),
#           a variable identifying the technical parameters input workbook used in the case ("pci_input_loc"),
#           and a vector of model fuel transport premiums and taxes ("params")

economic_inputs <- function(Unit_Type, pci_input_loc, params) {
  
  # 0. Set Input Paths ===============================================
  
  if(pci_input_loc == "[Default]") {
    Economic_Inputs <- paste(shared_inputs_folder, "Price and Cost Input Data.xlsx", sep = "/")
  }else{
    Economic_Inputs <- paste(run_inputs_folder, paste0(pci_input_loc), sep = "/")
  }
  
  # 1. Load Annually Updated Parameters =============================
  
  ## A. DA LBMPs ------------------------------------------------
  
  da_lbmps <- xlsx_import(Economic_Inputs, "DA LBMP - All Zones")
  
  da_lbmps = da_lbmps %>% filter(!is.na(Eastern.Date.Hour)) %>%
    mutate(Date = as_date(Eastern.Date.Hour), Hour = hour(Eastern.Date.Hour)) %>%
    rename(DateTime = Eastern.Date.Hour, DAM.LBMP = DAM.Zonal.LBMP, Zone = Zone.Name) %>%
    select(DateTime, Date, Hour, Zone, DAM.LBMP)
  
  ## B. RT LBMPs ----------------------------------------------------
  
  rt_lbmps <- xlsx_import(Economic_Inputs, "RT LBMP - All Zones")
  
  rt_lbmps <- rt_lbmps %>% filter(!is.na(Eastern.Date.Hour)) %>%
    mutate(Date = as_date(Eastern.Date.Hour), Hour = hour(Eastern.Date.Hour)) %>%
    rename(DateTime = Eastern.Date.Hour, RT.LBMP = TWI.Zonal.LBMP, Zone = Zone.Name) %>%
    select(DateTime, Date, Hour, Zone, RT.LBMP)
  
  ## C. Nonspin Reserve Prices (CT Only) -----------------------------
  ### i. DA NSR Prices
  
  if (Unit_Type == "CT") {
    da_anc <- xlsx_import(Economic_Inputs, "DA Ancillary - All Zones")
    
    da_anc <- da_anc %>% filter(!is.na(Eastern.Date.Hour)) %>%
      mutate(Date = as_date(Eastern.Date.Hour), Hour = hour(Eastern.Date.Hour)) %>%
      rename(DateTime = Eastern.Date.Hour, DA.NSR.10Mins = DAM.10.Min.Non.Sync, DA.30Mins = DAM.30.Min.Non.Sync, Zone = Zone.Name)
    da_anc <- da_anc %>% select(DateTime, Date, Hour, Zone, DA.NSR.10Mins, DA.30Mins)
    
    ### ii. RT NSR Prices
    
    rt_anc <- xlsx_import(Economic_Inputs, "RT Ancillary - All Zones")
    
    rt_anc <- rt_anc %>% filter(!is.na(Hour.Begin.Time.Stamp)) %>%
      mutate(Date = as_date(Hour.Begin.Time.Stamp), Hour = hour(Hour.Begin.Time.Stamp)) %>%
      rename(DateTime = Hour.Begin.Time.Stamp,
             RT.NSR.10Mins = ..10.Min.Non..Synchronous.Reserve....MWHr.,
             RT.SR.10Mins = ..10.Min.Spinning.Reserve..............MWHr.,
             RT.30Mins = ..30.Min.Operating.Reserve..............MWHr.)
    rt_anc <- rt_anc %>% select(DateTime, Date, Hour, Zone, RT.NSR.10Mins, RT.30Mins)
  }
  
  ## D. Generate list of all energy prices, apply Daylight Savings adjustment ------------------
  
  if(Unit_Type == "CT") {
    all_energy_prices <- list(da_lbmps, rt_lbmps, da_anc, rt_anc)
    names(all_energy_prices) <- c("da_lbmps", "rt_lbmps", "da_anc", "rt_anc")
  }else{
    all_energy_prices <- list(da_lbmps, rt_lbmps)
    names(all_energy_prices) <- c("da_lbmps", "rt_lbmps")
  }
  
  all_energy_prices <- lapply(all_energy_prices, dls_adjust)
  all_energy_prices <- lapply(all_energy_prices, set_daterange, start_date = start_date, end_date = end_date)
  
  ## E. Fuel Prices --------------------------------------------------
  ### i. Gas Prices
  snl_gas <- xlsx_import(Economic_Inputs, "SNL Gas")
  snl_gas[2:length(snl_gas)] <- apply(snl_gas[2:length(snl_gas)], 2, as.double) %>%
    apply(2, function (x) ifelse(is.na(x), 0, x)) 
  snl_gas <- snl_gas %>% filter(!is.na(Date)) %>% arrange(Date)
  
  ### ii. Oil Prices
  
  eia_oil <- xlsx_import(Economic_Inputs, "EIA Oil")
  eia_oil <- eia_oil %>% rename(Date = Day) %>% filter(!is.na(Date)) %>% arrange(Date)
  eia_oil$Date <- as_date(eia_oil$Date)
  
  all_fuel_prices <- list(snl_gas, eia_oil)
  names(all_fuel_prices) <- c("snl_gas", "eia_oil")
  
  ## F. Emissions Prices ----------------------------------------------
  
  ### i. RGGI Prices
  rggi_co2 <- xlsx_import(Economic_Inputs, "RGGI CO2")
  rggi_co2 <- rggi_co2 %>% filter(!is.na(Auction.Date)) %>% rename(CO2.Price = Auction.Clearing.Price) %>% arrange(Auction.Date)
  rggi_co2$Auction.Date <- as_date(rggi_co2$Auction.Date)
  
  ### ii. NOx Prices
  snl_nox <- xlsx_import(Economic_Inputs, "SNL NOx")
  snl_nox <- snl_nox %>% filter(!is.na(Date)) %>% 
    rename(Nox.Seasonal.Price = Seasonal.Nox, Nox.Annual.Price = Annual.Nox) %>%
    arrange(Date)
  snl_nox$Nox.Annual.Price <- as.double(snl_nox$Nox.Annual.Price)
  snl_nox$Nox.Seasonal.Price <- as.double(snl_nox$Nox.Seasonal.Price)
  
  ### iii. CSAPR SOx Prices
  snl_csapr_so2 <- xlsx_import(Economic_Inputs, "SNL CSAPR SO2")
  snl_csapr_so2 <- snl_csapr_so2 %>% filter(!is.na(Date)) %>% rename(So2.Price = Allowance.Price) %>% arrange(Date)
  
  ### iv. ARP SOx Prices
  snl_arp_so2 <- xlsx_import(Economic_Inputs, "SNL ARP SO2")
  snl_arp_so2 <- snl_arp_so2 %>% filter(!is.na(Date)) %>% rename(So2.Price = Allowance.Price) %>% arrange(Date)
  
  all_emissions_prices <- list(rggi_co2, snl_nox, snl_csapr_so2, snl_arp_so2)
  names(all_emissions_prices) <- c("rggi_co2", "snl_nox", "snl_csapr_so2", "snl_arp_so2")  
  
  
  ## G. RS1 ----------------------------------------------------------
  
  rs1 <- xlsx_import(Economic_Inputs, "NYISO RS1")
  rs1 <- rs1 %>% filter(!is.na(Year)) %>% arrange(Year)
  
  ## H. Modeling Dates -----------------------------------------------
  
  dates <- xlsx_import(Economic_Inputs, "Dates")
  dates <- dates %>%  filter(!is.na(Date), Date >= start_date & Date <= end_date)
  dates$Model.Year <- case_when(dates$Date < (start_date + years(1)) ~ 1,
                                dates$Date < (start_date + years(2)) ~ 2,
                                dates$Date < (start_date + years(3)) ~ 3,
                                TRUE ~ NA_real_)
  
  # 2. Load Universal DCR Parameters ------------------------------------------
  
  ## A. Fuel Transport Costs
  fuel_costs <- params %>% filter(Parameter != "Start Date") %>%
    rename(Item = Parameter, Cost = Value)
  fuel_costs$Cost <- as.numeric(fuel_costs$Cost)
  
  
  ## B. LOEs
  
  loe_wide <- xlsx_import(Economic_Inputs, "LOE")
  loe_wide <- loe_wide %>% mutate(DateTime = as.POSIXct(Date + hours(Hour) + seconds(Second))) %>%
    arrange(DateTime) %>%
    select(-Hour, -Second) %>%
    select(DateTime, everything())
  names(loe_wide)[c(5,6)] <- c("HUD VL", "N.Y.C.")
  names(loe_wide) <- c("DateTime", "Date", paste0(c("Zone."), names(select(loe_wide, CENTRL:LONGIL))))
  
  loe_vars_to_reshape <- names(select(loe_wide, Zone.CENTRL:Zone.LONGIL))
  loe_long <- as.data.frame(loe_wide) %>% reshape(direction = "long",
                                                  varying = 3:7,
                                                  v.names = c("LOE"),
                                                  timevar = c("Zone"),
                                                  times = gsub("Zone.", "", loe_vars_to_reshape),
                                                  idvar = c("DateTime"),
                                                  sep = ".") %>%
    filter(Date >= start_date & Date <= end_date) %>% select(-Date)
  
  prices_and_parameters <- list(all_energy_prices, all_emissions_prices, all_fuel_prices,
                                rs1, dates, fuel_costs, loe_long)
  names(prices_and_parameters) <- c("all_energy_prices", "all_emissions_prices", "all_fuel_prices",
                                    "rs1", "dates", "fuel_costs", "loe") 
  return(prices_and_parameters)
}

set_daterange <- function(df, start_date, end_date) {
  df <- filter(df, df$Date >= start_date & Date <= end_date)
  return(df)
}
