###################################################################
# NYISO DCR Reset 2020: Net EAS Revenues Model
# Module 1, part 3
#_________________________________________________________________;

# Script purpose 
#   This script includes a function that takes a set of technical parameters, prices/cost parameters, and case specifications and generates a single set of inputs for each unit in a case
#   Inputs: a set of prices & cost parameters produced by the function in script 1.1 ("Economic_inputs"); 
#           a set of unit technical parameters produced by the function in script 1.2 ("Engineering_Inputs");
#           a toggle identifying whether the case involves CTs or CCs ("Unit_Type")
#           the first date in the three-year period examined in the run ("start_date")

#### 0. Setup ----------------------------------------------------

# Import data and define constants
input_merge <- function(Economic_inputs, Engineering_Units, Engineering_Inputs, Unit_Type, Zone_C_Hub, start_date) {
  end_date <- ymd(paste0(year(start_date)+3, "08", "31"))
  
  ## 1. Emissions Prices -----------------------------------------
  all_emissions_prices <- Economic_inputs$all_emissions_prices
  
  ### Merge together emissions prices
  em_merge_proc <- full_join(Economic_inputs$dates,all_emissions_prices$rggi_co2, by = c("Date" = "Auction.Date")) %>%
    full_join(all_emissions_prices$snl_nox, by = "Date") %>%
    full_join(all_emissions_prices$snl_csapr_so2, by = "Date") %>%
    left_join(all_emissions_prices$snl_arp_so2, by = "Date") %>%
    rename(so2.csapr.price = So2.Price.x, so2.arp.price = So2.Price.y) %>%
    select(Date, Model.Year, CO2.Price, Nox.Annual.Price, Nox.Seasonal.Price, so2.csapr.price, so2.arp.price) %>%
    arrange(Date)
  em_merge_proc$Nox.Annual.Price <- ifelse(em_merge_proc$Nox.Annual.Price == "-", NA, em_merge_proc$Nox.Annual.Price)
  em_merge_proc$Nox.Seasonal.Price <- ifelse(em_merge_proc$Nox.Seasonal.Price == "-", NA, em_merge_proc$Nox.Seasonal.Price)
  
  ### Carry forward previous emissions price when missing
  merged_emissions <- em_merge_proc %>% fill(CO2.Price:so2.arp.price)
  
  ### Remove dates that are out of range
  merged_emissions <- merged_emissions %>% subset(Date >= start_date & Date <= end_date)
  
  ## 2. Fuel Prices --------------------------------------------
  
  ### Calculate prices for each zone
  all_fuel_prices <- Economic_inputs$all_fuel_prices
  gas_matrix_sheet <- "Gas_Matrix"
  gas_hub_table <- xlsx_import(paste0(shared_inputs_folder, "General Model Parameters.xlsx"), gas_matrix_sheet) %>% as.data.frame(lapply(as.numeric))
  gas_hub_table_multiply <- select(gas_hub_table, -Hub) %>% as.matrix()
  if(Zone_C_Hub == "Niagara / TGP Z4 200L"){
    all_fuel_prices_multiply <- mutate(all_fuel_prices$snl_gas, TGP_Niagara_Blend = ifelse(month(Date) %in% c(1, 2, 3, 12), Niagara_Price, TGP_Z4_200L_Price))  %>%
      select(-Date, -TGP_Z4_200L_Price, -Niagara_Price)
  }else{
    all_fuel_prices_multiply <- select(all_fuel_prices$snl_gas, -Date, -Niagara_Price) 
  }
  all_fuel_prices_multiply <- as.matrix(all_fuel_prices_multiply)
  
  all_fuel_prices_zones <- all_fuel_prices_multiply %*% gas_hub_table_multiply
  all_fuel_prices_zones <- as_tibble(cbind(all_fuel_prices$snl_gas$Date, 
                                           all_fuel_prices_zones))
  names(all_fuel_prices_zones) <- c("Date", paste0("Gas.Price.Zone.", names(gas_hub_table)[2:7]))
  all_fuel_prices_zones$Date <- as_date(all_fuel_prices_zones$Date)
  
  merged_fuel <- full_join(Economic_inputs$dates, all_fuel_prices_zones, by = "Date") %>%
    full_join(all_fuel_prices$eia_oil, by = "Date") %>%
    rename(ULSD.Price = U.S..No.2.Diesel.Ultra.Low.Sulfur.Less.than.15.ppm.Retail.Sales.by.Refiners.Dollars.per.Gallon) %>%
    select(-Model.Year) %>%
    ### Fill fuel prices for missing dates with prices from the next available date
    arrange(desc(Date)) %>%
    fill(Gas.Price.Zone.C:ULSD.Price) %>%
    arrange(Date) %>%
    subset(Date >= start_date & Date <= end_date)
  merged_fuel$ULSD.Price <- merged_fuel$ULSD.Price * 7.2790269396787
  ### Note: this conversion factor takes the original $/gallon and turns it into $/MMBtu per EIA conversion factor of
  ###       137,381 BTU/gallon
  
  
  ## 3. Combine All Datasets  --------------------------------------------
  
  ### Merge in economic inputs
  
  all_energy_prices <- Economic_inputs$all_energy_prices
  merge_base <- Economic_inputs$all_energy_prices$da_lbmps %>% mutate(Area = case_when(Zone == "CENTRL" ~ "C",
                                                                                       Zone == "CAPITL" ~ "F",
                                                                                       Zone == "HUD VL" ~ "G1",
                                                                                       Zone == "N.Y.C." ~ "J",
                                                                                       Zone == "LONGIL" ~ "K"),
                                                                      Capacity_Zone = case_when(Zone == "CENTRL" | Zone == "CAPITL" ~ "ROS",
                                                                                                Zone == "HUD VL" ~ "HV",
                                                                                                Zone == "N.Y.C." ~ "NYC",
                                                                                                Zone == "LONGIL" ~ "LI")) %>%
    filter(!is.na(Area))
  merge_hv_duplicates <- merge_base[which(merge_base$Zone == "HUD VL"), ]
  merge_hv_duplicates$Area <- "G2"
  merge_base <- merge_base %>% rbind(merge_hv_duplicates) %>%
    arrange(DateTime, Zone)
  
  merge_base_nounits <- merge_base %>% full_join(Economic_inputs$dates, by = c("Date")) %>%
    left_join(all_energy_prices$rt_lbmps) %>%
    subset(Date >= start_date & Date <= end_date)
  
  if (Unit_Type == "CT") {
    merge_base_nounits <- merge_base_nounits %>% 
      left_join(all_energy_prices$da_anc) %>%
      fill(DA.NSR.10Mins, DA.30Mins) %>%
      left_join(all_energy_prices$rt_anc) 
  }
  
  merge_base_nounits <- merge_base_nounits %>% left_join(Economic_inputs$loe)
  
  merge_base_nounits <- merge_base_nounits %>% left_join(merged_fuel) %>%
    left_join(merged_emissions) %>%
    mutate(Year = year(Date)) %>%
    left_join(Economic_inputs$rs1) %>%
    select(-Year) %>%
    select(contains("Date"), Hour, Model.Year, Area, contains("Zone"), contains("DAM"), contains("RT"), contains("Gas.Price"), 
           everything())  
  
  
  ### Merge in unit technical parameters
  
  for (i in 1:(length(Engineering_Units))) {
    unit_eval <- Engineering_Units[i]
    Engineering_Inputs_eval <- Engineering_Inputs[[i]]
    name_extract_set <- select(Engineering_Inputs_eval, -AREA, -TECH_PARAMS)
    names_with_units <- names(name_extract_set)
    names(Engineering_Inputs_eval)[3:(length(name_extract_set) + 2)] <- substr(names_with_units, 1, nchar(names_with_units) - (nchar(unit_eval) + 1))
    if (i == 1) {
      merge_base_forfunc <- merge_base_nounits
    }else{
      merge_base_forfunc <- merge_base_full
    }
    
    
    merge_base_full <- merge_base_forfunc %>% left_join(Engineering_Inputs_eval, by = c("Area" = "AREA")) %>%
      mutate(summer_flag = ifelse(month(Date) %in% c(5:10), 1, 0)) %>%
      mutate(CAP = ifelse(summer_flag == 1, CAP_SUM, CAP_WIN),
             HR = ifelse(summer_flag == 1, HR_SUM / 1000, HR_WIN / 1000),
             CO2_NG = ifelse(summer_flag == 1, (CO2_NG_SUM / CAP_SUM) / 2000, (CO2_NG_WIN / CAP_WIN) / 2000),
             CO2_OIL = ifelse(summer_flag == 1, (CO2_OIL_SUM / CAP_SUM) / 2000, (CO2_OIL_WIN / CAP_WIN) / 2000),
             NOX_NG = ifelse(summer_flag == 1, (NOX_NG_SUM / CAP_SUM) / 2000, (NOX_NG_WIN / CAP_WIN) / 2000),
             NOX_OIL = ifelse(summer_flag == 1, (NOX_OIL_SUM / CAP_SUM) / 2000, (NOX_OIL_WIN / CAP_WIN) / 2000),
             SO2_NG = ifelse(summer_flag == 1, (SO2_NG_SUM / CAP_SUM) / 2000, (SO2_NG_WIN / CAP_WIN) / 2000),
             SO2_OIL = ifelse(summer_flag == 1, (SO2_OIL_SUM / CAP_SUM) / 2000, (SO2_OIL_WIN / CAP_WIN) / 2000))
    if(Unit_Type == "CC") {
      merge_base_full <- merge_base_full %>% mutate(MIN_CAP_RATIO = ifelse(summer_flag == 1, CAP_SUM_MIN / CAP_SUM, CAP_WIN_MIN / CAP_WIN),
                                                    MIN_HR = ifelse(summer_flag == 1, HR_SUM_MIN / 1000, HR_WIN_MIN / 1000),
                                                    CO2_NG_MIN = ifelse(summer_flag == 1, (CO2_NG_SUM_MIN / CAP_SUM) / 2000, (CO2_NG_WIN_MIN / CAP_WIN) / 2000),
                                                    CO2_OIL_MIN = ifelse(summer_flag == 1, (CO2_OIL_SUM_MIN / CAP_SUM) / 2000, (CO2_OIL_WIN_MIN / CAP_WIN) / 2000),
                                                    NOX_NG_MIN = ifelse(summer_flag == 1, (NOX_NG_SUM_MIN / CAP_SUM) / 2000, (NOX_NG_WIN_MIN / CAP_WIN) / 2000),
                                                    NOX_OIL_MIN = ifelse(summer_flag == 1, (NOX_OIL_SUM_MIN / CAP_SUM) / 2000, (NOX_OIL_WIN_MIN / CAP_WIN) / 2000),
                                                    SO2_NG_MIN = ifelse(summer_flag == 1, (SO2_NG_SUM_MIN / CAP_SUM) / 2000, (SO2_NG_WIN_MIN / CAP_WIN) / 2000),
                                                    SO2_OIL_MIN = ifelse(summer_flag == 1, (SO2_OIL_SUM_MIN / CAP_SUM) / 2000, (SO2_OIL_WIN_MIN / CAP_WIN) / 2000))
    }
    merge_base_full <- select(merge_base_full, -(NOX_NG_SUM:CO2_OIL_WIN), -(HR_SUM:HR_WIN), -(CAP_SUM:CAP_WIN), -summer_flag, -TECH_PARAMS)
    if (Unit_Type == "CC") {
      merge_base_full <- select(merge_base_full, -(HR_SUM_MIN:CAP_WIN_MIN))
    }
    names_with_units <- names(merge_base_full[(which(names(merge_base_full) == "VOC_NG")):length(merge_base_full)])
    names(merge_base_full)[(which(names(merge_base_full) == "VOC_NG")):length(merge_base_full)] <- paste0(names_with_units, "__", toupper(unit_eval))
    
  }
  
  
  ## 4. Set Proper Gas in Each Zone ----------------------------------
  
  fuel_costs <- as_tibble(t(Economic_inputs[["fuel_costs"]]))
  names(fuel_costs) <- gsub(" ", ".", fuel_costs[1,])
  fuel_costs <- slice(fuel_costs, 2)
  fuel_costs <- fuel_costs %>% lapply(as.numeric) %>% as_tibble()
  
  merge_base_full_gas <- merge_base_full %>% mutate(DAM.Gas.Price = 
                                                      case_when(Area == "C" ~ Gas.Price.Zone.C + fuel_costs$ROS.Fuel.TransCost,
                                                                Area == "F" ~ Gas.Price.Zone.F + fuel_costs$ROS.Fuel.TransCost,
                                                                Area == "G1" ~ Gas.Price.Zone.G1 + fuel_costs$HudVl.Fuel.TransCost,
                                                                Area == "G2" ~ Gas.Price.Zone.G2 + fuel_costs$HudVl.Fuel.TransCost,
                                                                Area == "J" ~ Gas.Price.Zone.J * fuel_costs$NYC.Gas.Tax + fuel_costs$NYC.Fuel.TransCost,
                                                                Area == "K" ~ Gas.Price.Zone.K * fuel_costs$LI.Gas.Tax + fuel_costs$LI.Fuel.TransCost),
                                                    RTD.Gas.Price.Buy =
                                                      case_when(Area == "C" ~ Gas.Price.Zone.C * fuel_costs$ROS.Fuel.Premium + fuel_costs$ROS.Fuel.TransCost,
                                                                Area == "F" ~ Gas.Price.Zone.F * fuel_costs$ROS.Fuel.Premium + fuel_costs$ROS.Fuel.TransCost,
                                                                Area == "G1" ~ Gas.Price.Zone.G1 * fuel_costs$HudVL.Fuel.Premium + fuel_costs$HudVl.Fuel.TransCost,
                                                                Area == "G2" ~ Gas.Price.Zone.G2 * fuel_costs$HudVL.Fuel.Premium + fuel_costs$HudVl.Fuel.TransCost,
                                                                Area == "J" ~ Gas.Price.Zone.J * fuel_costs$NYC.Fuel.Premium * fuel_costs$NYC.Gas.Tax
                                                                + fuel_costs$NYC.Fuel.TransCost,
                                                                Area == "K" ~ Gas.Price.Zone.K * fuel_costs$LI.Fuel.Premium * fuel_costs$LI.Gas.Tax
                                                                + fuel_costs$LI.Fuel.TransCost),
                                                    RTD.Gas.Price.Sell =
                                                      case_when(Area == "C" ~ Gas.Price.Zone.C * fuel_costs$ROS.Fuel.Disc + fuel_costs$ROS.Fuel.TransCost,
                                                                Area == "F" ~ Gas.Price.Zone.F * fuel_costs$ROS.Fuel.Disc + fuel_costs$ROS.Fuel.TransCost,
                                                                Area == "G1" ~ Gas.Price.Zone.G1 * fuel_costs$HudVl.Fuel.Disc + fuel_costs$HudVl.Fuel.TransCost,
                                                                Area == "G2" ~ Gas.Price.Zone.G2 * fuel_costs$HudVl.Fuel.Disc + fuel_costs$HudVl.Fuel.TransCost,
                                                                Area == "J" ~ Gas.Price.Zone.J * fuel_costs$NYC.Fuel.Disc * fuel_costs$NYC.Gas.Tax 
                                                                + fuel_costs$NYC.Fuel.TransCost,
                                                                Area == "K" ~ Gas.Price.Zone.K * fuel_costs$LI.Fuel.Disc * fuel_costs$LI.Gas.Tax
                                                                + fuel_costs$LI.Fuel.TransCost)) %>%
    select(-(Gas.Price.Zone.C:Gas.Price.Zone.K))
  
  merge_base_full_gas$ULSD.Price <- case_when(merge_base_full_gas$Area %in% c("C", "F") ~ merge_base_full_gas$ULSD.Price + fuel_costs$ROS.ULSD.TransCost,
                                              merge_base_full_gas$Area %in% c("G1", "G2") ~ merge_base_full_gas$ULSD.Price + fuel_costs$HudVl.ULSD.TransCost,
                                              merge_base_full_gas$Area == "J" ~ merge_base_full_gas$ULSD.Price * fuel_costs$NYC.Sales.Tax 
                                              + fuel_costs$NYC.ULSD.TransCost,
                                              merge_base_full_gas$Area == "K" ~ merge_base_full_gas$ULSD.Price + fuel_costs$LI.ULSD.TransCost)
  
  
  ## 5. Final Changes & Checks ------------------------------------------------
  
  merge_base_near_final <- arrange(merge_base_full_gas, Area, DateTime)
  
  ### Address missing emissions, rates, and  values
  vars_to_check <- c("CO2.Price", "Nox.Annual.Price", "Nox.Seasonal.Price", "so2.csapr.price", "so2.arp.price", "Rate")
  merge_base_near_final[,which(names(merge_base_near_final) %in% vars_to_check)] <- lapply(merge_base_near_final[, which(names(merge_base_near_final) %in% vars_to_check)], function(x) ifelse(is.na(x), 0, x))
  merge_base_near_final$LOE <- ifelse(is.na(merge_base_near_final$LOE), 1, merge_base_near_final$LOE)
  
  ### Set Nox Prices
  merge_base_near_final <- merge_base_near_final %>% mutate(Nox.Price = 
                                                              ifelse(month(Date) %in% c(5:9), Nox.Seasonal.Price + Nox.Annual.Price, Nox.Annual.Price)) %>%
    select(-Nox.Annual.Price, -Nox.Seasonal.Price)
  
  ### Set Sox Prices
  merge_base_near_final <- merge_base_near_final %>% mutate(So2.price =
                                                              ifelse(year(Date) %in% c(2010:2014), 2*so2.csapr.price + so2.arp.price, so2.csapr.price + so2.arp.price)) %>%
    select(-so2.csapr.price, -so2.arp.price)
  
  eas_func_base <- merge_base_near_final
  
  
  return(eas_func_base)
}
