##################################################################
# NYISO DCR Reset 2020: Net EAS Revenues Model
# Module 2, Part 1
#_________________________________________________________________

# Script purpose 
#   This script includes a function that takes a case and calculates potential profits in each hour for each unit within that case.
#   Inputs: a case number ("case_to_eval") and a full set of inputs (prices, cost parameters, and unit technical parameters) produced by the function in scripts 1.3 ("inputs")


#### 0. Prepare inputs: select case and toggle parameters----------------------------------------------------------------------

eval_case_base <- function(inputs, case_to_eval) {
  case_base <- inputs[[paste0("Case_", case_to_eval)]]
  case_units_toggles <- case_base[[1]]
  toggle_list <<- case_units_toggles[[2]]
  case_inputs <- case_base[[2]]
  
  if (toggle_list$Zones == "All Zones") {
    zone_vec <- c("C", "F", "G1", "G2", "J", "K")
  }else if(toggle_list$Zones == "Gas-Only Zones") {
    zone_vec <- c("C", "F", "G1", "G2")
  }else if(toggle_list$Zones == "No-SCR Zones") {
    zone_vec <- c("C", "F", "G1")
  }else if(toggle_list$Zones == "ROS") {
    zone_vec <- c("C", "F")
  }else if(toggle_list$Zones == "G-K Zones"){
    zone_vec <- c("G1", "G2", "J", "K") 
  }else if(toggle_list$Zones == "G-J Zones"){
    zone_vec <- c("G1", "G2", "J")
  }else{
    zone_vec <- c(toggle_list$Zones)
  }
  
  ## Gas Only Toggle: address inconsistencies between zone selection and this toggle
  if(toggle_list$Gas_Only_Toggle == "Yes") {
    case_inputs$ULSD.Price = 99999
    if ("J" %in% zone_vec | "K" %in% zone_vec) {
      warning(paste("For Case", case_to_eval, ", gas-only units cannot run in Zones J and K. These zones will not be evaluated."))
      case_inputs <- case_inputs %>% filter(Area %in% c("C", "F", "G1", "G2"))  
    }
  }
  
  ## No-SCR Toggle: address inconsistencies between zone selection and this toggle
  if(toggle_list$SCR_Toggle == "No" & ("J" %in% zone_vec | "K" %in% zone_vec | "G2" %in% zone_vec)) {
    warning(paste("For Case", paste0(case_to_eval, ","), "be aware that no-SCR units cannot run in severe non-attainment zones (Zones J and K, as well as the Rockland portion of Zone G2). These zones will not be evaluated."))
    case_inputs <- case_inputs %>% filter(Area %in% c("C", "F", "G1"))
  }
  
  ## Drop zones not selected in the zone_vec
  case_inputs <- case_inputs %>% filter(Area %in% zone_vec)
  techparams_input_loc <- case_units_toggles$Toggles$Technical_Inputs_Filename
  
  unit_list <- case_units_toggles$Unit_List
  
  ## Set function to evaluate input base for each unit
  eval_input_base <- function(unit_to_eval, techparams_input_loc, case_to_eval) {
    
    ### Select subset of units to evaluate
    unit_suffix <- paste0("__", toupper(unit_to_eval))
    parameter_subset_start <- paste0("VOC_NG", unit_suffix)
    parameter_subset_end <- ifelse(case_units_toggles$Toggles$Unit_Type == "CC", paste0("SO2_OIL_MIN", unit_suffix), paste0("SO2_OIL", unit_suffix))
    case_inputs_subset <- select(case_inputs, DateTime:Rate, DAM.Gas.Price:So2.price, parameter_subset_start:parameter_subset_end)
    
    ### Strip suffixes
    names(case_inputs_subset)[(which(names(case_inputs_subset) == "Rate") + 1):length(names(case_inputs_subset))] <-
      gsub(unit_suffix, "", names(case_inputs_subset)[(which(names(case_inputs_subset) == "Rate") + 1):length(names(case_inputs_subset))])
    
    ### 1. Select  use of oil or gas ----------------------------------------------------------------
    #### Generate costs of using oil vs. gas
    case_inputs_subset <- case_inputs_subset %>% mutate(DAM.Gas.Fuel.Cost = HR * DAM.Gas.Price,
                                                        RTD.Gas.Fuel.Cost = HR * RTD.Gas.Price.Buy,
                                                        Gas.Emissions.Cost = CO2_NG * CO2.Price + NOX_NG * Nox.Price + SO2_NG * So2.price,
                                                        DAM.Gas.Cost = DAM.Gas.Fuel.Cost + Gas.Emissions.Cost + VOC_NG + Rate,
                                                        RTD.Gas.Cost = RTD.Gas.Fuel.Cost + VOC_NG + Rate + Gas.Emissions.Cost,
                                                        Oil.Fuel.Cost = HR * ULSD.Price,
                                                        Oil.Emissions.Cost = CO2_OIL * CO2.Price + NOX_OIL * Nox.Price + SO2_OIL * So2.price,
                                                        Oil.Cost = Oil.Fuel.Cost + VOC_OIL + Rate + Oil.Emissions.Cost)
    #### For aeroderivatives, major maintenance costs accumulate on a per-hr of operation basis rather than on a per-start basis. 
    #### Per-start MM costs are 0 and per-hour MM costs are instead incorporated here.
    #### Note: CCs make operational decisions based upon $/start major maintenance costs but incur costs on a $/hr basis.
    #### Final accounting in 2.3 will incorporate $/hr major maintenance costs.
    if(unit_to_eval == "aero") {
      case_inputs_subset <- case_inputs_subset %>% mutate(DAM.Gas.Cost = DAM.Gas.Cost + MM_COST / CAP,
                                                          RTD.Gas.Cost = RTD.Gas.Cost + MM_COST / CAP,
                                                          Oil.Cost = Oil.Cost + MM_COST  / CAP)
    }
    case_inputs_subset <- select(case_inputs_subset, -(DAM.Gas.Fuel.Cost:RTD.Gas.Fuel.Cost), -Oil.Fuel.Cost)
    
    #### For CC units, generate minimum load operation costs
    #### Note: these costs do not include major maintenance costs, as CCs make operational decisions based upon $/start major maintenance costs. 
    #### Final accounting in 2.4 will incorporate $/hr major maintenance costs.
    if(case_units_toggles$Toggles$Unit_Type == "CC") {
      case_inputs_subset <- case_inputs_subset %>% mutate(DAM.Gas.Fuel.Min.Cost = MIN_HR * DAM.Gas.Price,
                                                          RTD.Gas.Fuel.Min.Cost = MIN_HR * RTD.Gas.Price.Buy,
                                                          Gas.Emissions.Min.Cost = CO2_NG_MIN * CO2.Price + NOX_NG_MIN * Nox.Price + SO2_NG_MIN * So2.price,
                                                          DAM.Gas.Min.Cost = MIN_CAP_RATIO * (DAM.Gas.Fuel.Min.Cost + VOC_NG + Rate + Gas.Emissions.Min.Cost),
                                                          RTD.Gas.Min.Cost = MIN_CAP_RATIO * (RTD.Gas.Fuel.Min.Cost + VOC_NG + Rate + Gas.Emissions.Min.Cost),
                                                          Oil.Fuel.Min.Cost = MIN_HR * ULSD.Price,
                                                          Oil.Emissions.Min.Cost = CO2_OIL_MIN * CO2.Price + NOX_OIL_MIN * Nox.Price + SO2_OIL_MIN * So2.price,
                                                          Oil.Min.Cost = MIN_CAP_RATIO * (Oil.Fuel.Min.Cost + VOC_OIL + Rate + Oil.Emissions.Min.Cost)) %>%
        select(-DAM.Gas.Fuel.Min.Cost, -RTD.Gas.Fuel.Min.Cost, -Oil.Fuel.Min.Cost)
    }
    
    #### Select oil or gas based on the day-ahead gas price and update fuel cost parameters accordingly
    case_inputs_subset <- case_inputs_subset %>% 
      mutate(Economic.Fuel = "Gas",
             DAM.Economic.Fuel.Price = DAM.Gas.Price,
             RTD.Economic.Fuel.Price = RTD.Gas.Price.Buy,
             DAM.Cost = DAM.Gas.Cost,
             RTD.Cost = RTD.Gas.Cost) %>% 
      as.data.table()
    
    case_inputs_subset[Oil.Cost < DAM.Gas.Cost & !is.na(ULSD.Price), ':=' (Economic.Fuel = "Oil",
                                                                           DAM.Economic.Fuel.Price = ULSD.Price,
                                                                           RTD.Economic.Fuel.Price = ULSD.Price,
                                                                           DAM.Cost = Oil.Cost,
                                                                           RTD.Cost = Oil.Cost)]
    if(case_units_toggles$Toggles$Unit_Type == "CC"){
      case_inputs_subset$DAM.Min.Cost = ifelse(case_inputs_subset$Economic.Fuel == "Oil",
                                               case_inputs_subset$Oil.Min.Cost,
                                               case_inputs_subset$DAM.Gas.Min.Cost)
      case_inputs_subset$RTD.Min.Cost = ifelse(case_inputs_subset$Economic.Fuel == "Oil",
                                               case_inputs_subset$Oil.Min.Cost,
                                               case_inputs_subset$RTD.Gas.Min.Cost)
    }
    
    ### 2. Calculate Potential Profits in Each Hour ------------------------------------
    
    #### Calculate fuel losses if unit buys out day-ahead fuel commitment and has to resell the gas
    case_inputs_subset$E.Buyout.Fuel.Loss.per.MMBtu <- ifelse(case_inputs_subset$Economic.Fuel == "Gas",
                                                              case_inputs_subset$DAM.Gas.Price - case_inputs_subset$RTD.Gas.Price.Sell,
                                                              0)
    case_inputs_subset$E.Buyout.Fuel.Loss.per.MWh <- case_inputs_subset$E.Buyout.Fuel.Loss * case_inputs_subset$HR 
    
    #### Calculate reserves profits depending on whether the unit is seeking 30 minute or 10 minute non-spinning reserve provision
    #### Note:  1.  This is not calculated for CC units, as they don't provide non-spinning reserves
    ####        2.  In order to commit to providing day-ahead reserves, the unit has to secure fuel. Thus, in calculating DAM reserve profits,
    ####              we approximate the cost of providing DAM reserves as the cost of selling fuel back to the market.
    if(case_units_toggles$Toggles$Unit_Type == "CT"){
      source_wb <- ifelse(techparams_input_loc == "[Default]",
                          paste0(shared_inputs_folder, "Unit Technical Parameters.xlsx"),
                          paste0(run_inputs_folder, techparams_input_loc))
      ns_parameters <- xlsx_import(source_wb, "gen_params_CTs_reserves")
      ns_parameters_subset <- ns_parameters %>% select(contains(paste0("_", unit_to_eval)))
      case_inputs_subset$DAM.Reserve.Cost <- ifelse(case_inputs_subset$Area %in% c("G1", "G2", "J", "K") & case_inputs_subset$Economic.Fuel == "Gas", 2, case_inputs_subset$E.Buyout.Fuel.Loss.per.MWh)
      if(ns_parameters_subset[,which(names(ns_parameters_subset) == paste0("ns10min_", unit_to_eval))] == "Yes"){
        case_inputs_pot_profits <- case_inputs_subset %>% mutate(DAM.Hourly.Reserve.Profit = LOE * DA.NSR.10Mins - DAM.Reserve.Cost,
                                                                 RTD.Hourly.Reserve.Price = LOE * RT.NSR.10Mins) # Just the price here, costs not yet included
      }else{
        case_inputs_pot_profits <- case_inputs_subset %>% mutate(DAM.Hourly.Reserve.Profit = LOE * DA.30Mins - DAM.Reserve.Cost,
                                                                 RTD.Hourly.Reserve.Price = LOE * RT.30Mins)
      }
    }else{
      case_inputs_pot_profits <- case_inputs_subset
    }
    
    
    #### Calculate hourly energy profit and total cost pre-startup costs by zone (hard cap at $4,000/MWh, per transmission shortage cost max pricing)
    case_inputs_pot_profits <- case_inputs_pot_profits %>% mutate(DAM.Hourly.Energy.Profit = pmin(LOE * DAM.LBMP, 4000) - DAM.Cost,
                                                                  RTD.Hourly.Energy.Profit = pmin(LOE * RT.LBMP, 4000) - RTD.Cost)
    if(case_units_toggles$Toggles$Unit_Type == "CC") {
      case_inputs_pot_profits <- case_inputs_pot_profits %>% mutate(DAM.Hourly.Energy.Min.Profit = MIN_CAP_RATIO * pmin(LOE * DAM.LBMP, 4000) - DAM.Min.Cost,
                                                                    RTD.Hourly.Energy.Min.Profit = MIN_CAP_RATIO * pmin(LOE * RT.LBMP, 4000) - RTD.Min.Cost)
    }
    
    #### Calculate startup costs if a unit starts in a given hour
    #### Note: CC units use major maintenance costs on a $/start basis for decisions but a $/hr basis for actual profits. Profits are adjusted in 2.3 to reflect this.
    case_inputs_pot_profits <- case_inputs_pot_profits %>% mutate(DAM.Startup.Fuel.Cost = (STARTUP_FUEL * DAM.Economic.Fuel.Price) / CAP,
                                                                  DAM.Startup.MM.Cost = STARTUP_MM / CAP,
                                                                  DAM.Startup.Cost = DAM.Startup.Fuel.Cost + DAM.Startup.MM.Cost,
                                                                  RTD.Startup.Fuel.Cost = (STARTUP_FUEL * RTD.Economic.Fuel.Price) / CAP,
                                                                  RTD.Startup.MM.Cost= STARTUP_MM / CAP,
                                                                  RTD.Startup.Cost = RTD.Startup.Fuel.Cost + RTD.Startup.MM.Cost,
                                                                  E.Buyout.Startup.Fuel.Losses = (STARTUP_FUEL * E.Buyout.Fuel.Loss.per.MMBtu) / CAP) %>%
      select(-(DAM.Startup.Fuel.Cost), -(RTD.Startup.Fuel.Cost))
    return(case_inputs_pot_profits)
  }
  
  
  input_base_pot_profits <- lapply(unit_list, eval_input_base, techparams_input_loc = techparams_input_loc, case_to_eval = case_to_eval)
  names(input_base_pot_profits) <- paste0("Unit_", toupper(unit_list))
  base_pot_profits <- list(case_units_toggles, input_base_pot_profits)
  names(base_pot_profits) <- c("Units_and_Toggles", "Base_Potential_Profits")
  return(base_pot_profits)
}
