###################################################################
# NYISO DCR Reset 2020: Net EAS Revenues Model
# Module 2, Part 2
#_________________________________________________________________

# Script purpose 
#   This script includes a set of nested functions that takes a case and determines the day-ahead commitments for all units in that case, as well as real-time energy buyouts.
#   Inputs: a case number ("case_to_eval") and a full set of model inputs and potential hourly profits for all units in that case produced by the function in script 2.1 ("inputs") 

set_dam_commits <- function(inputs, case_to_eval) {
  case_base <- inputs[[paste0("Case_", case_to_eval)]]
  case_units_toggles <- case_base[[1]]
  case_inputs <- case_base[[2]]
  unit_list <- case_units_toggles$Unit_List
  
  dam_rt_commits <- function(unit_to_eval) {

    # 0. Set proper dataset to evaluate ---------------------------------
    
    input_set_name <- paste0("Unit_", toupper(unit_to_eval))
    case_set <- case_inputs[[which(names(case_inputs) == input_set_name)]]
    
    # 1. Calculate Potential Energy Profits -----------------------------
    
    case_set <- case_set %>% arrange(Area, DateTime) %>%
      group_by(Area) %>%
      mutate(DAM.Potential.Energy.Profit = ifelse(DAM.Hourly.Energy.Profit > 0, DAM.Hourly.Energy.Profit, NA))
    
    area_vec <- unique(case_set$Area)
    case_set_base <- case_set %>% filter(NA)
    calc_pot_en_profits <- function(area_to_eval) {
      case_set_slice <- case_set %>% filter(Area == area_to_eval)
      case_set_slice$DAM.Potential.Energy.Profit <- ifelse(!is.na(case_set_slice$DAM.Potential.Energy.Profit) & is.na(lag(case_set_slice$DAM.Potential.Energy.Profit)),
                                                           case_set_slice$DAM.Potential.Energy.Profit - case_set_slice$DAM.Startup.Cost,
                                                           case_set_slice$DAM.Potential.Energy.Profit)
      case_set_base <<- bind_rows(case_set_base, case_set_slice)
      return(case_set_base)
    }
    case_set <- lapply(area_vec, calc_pot_en_profits)[[length(area_vec)]]
    
    
    # 2. Generate a Potential Block Number for Each Segment ------------------
    
    case_set_pot_block_starts <- case_set %>% filter(!is.na(DAM.Potential.Energy.Profit) & is.na(lag(DAM.Potential.Energy.Profit))) %>%
      mutate(DAM.Potential.Block.Number = c(1:n()))
    
    case_set <- case_set %>% left_join(case_set_pot_block_starts) %>%
      fill("DAM.Potential.Block.Number")
    
    case_set$DAM.Potential.Block.Number <- ifelse(is.na(case_set$DAM.Potential.Energy.Profit), NA, case_set$DAM.Potential.Block.Number)
    
    # 3. Ensure that aeroderivative units can recover costs within an hour ---------
    ## Here, we examine each potential energy dispatch block for aeroderivative units to see whether they can recover their costs with their first hour of operation.
    ## Startup costs are already incorporated for the first hour in each potential energy dispatch block.
    ## But if net profits (including startup costs) are negative, the unit won't operate. So we must see whether it will operate in the subsequent hour.
    ## In that subsequent hour, we need to net out startup costs again; if net profits are as a result negative, we should once again consider the next hour.
    ## After doing these checks, we reset block numbers to account for blocks that were deleted.
    
    if(unit_to_eval %in% c("aero")) {
      print(paste0("Addressing startup costs for ", unit_to_eval))
      case_set$DAM.Block.Energy.Profit <- ifelse(case_set$DAM.Potential.Energy.Profit > 0, case_set$DAM.Potential.Energy.Profit, NA)
      replace_set <- filter(case_set, NA)
      for (j in unique(case_set$Area)) {
        for (i in 1:max(case_set[which(case_set$Area == j),]$DAM.Potential.Block.Number, na.rm = TRUE)) {
          case_slice <- filter(case_set, DAM.Potential.Block.Number == i, Area == j) 
          index <- 1
          while ((is.na(case_slice[index, "DAM.Block.Energy.Profit"]))) {
            index <- index + 1
            if (index > nrow(case_slice)) {
              break
            }
            case_slice[index,]$DAM.Block.Energy.Profit <- ifelse((case_slice[index,]$DAM.Potential.Energy.Profit - case_slice[index,]$DAM.Startup.Cost) > 0,
                                                                 case_slice[index,]$DAM.Potential.Energy.Profit - case_slice[index,]$DAM.Startup.Cost,
                                                                 as.numeric(NA))
          }
          if (index > nrow(case_slice)) {
            case_set[which(case_set$DAM.Potential.Block.Number == i & case_set$Area == j),]$DAM.Block.Energy.Profit <- as.numeric(NA)
          }else{
            replace_set <- bind_rows(replace_set, case_slice)
          }
        }
      }
      case_set <- case_set %>% select(-DAM.Block.Energy.Profit) %>% left_join(replace_set)
      
      ## Reset Block Numbers
      
      print(paste0("Resetting block numbers for ", unit_to_eval))
      
      case_set_block_starts <- case_set %>% filter(!is.na(DAM.Block.Energy.Profit) & is.na(lag(DAM.Block.Energy.Profit))) %>%
        mutate(DAM.Block.Number = c(1:n()))
      
      case_set <- case_set %>% left_join(case_set_block_starts) %>%
        fill("DAM.Block.Number")
      
      case_set$DAM.Block.Number <- ifelse(is.na(case_set$DAM.Block.Energy.Profit), NA, case_set$DAM.Block.Number)
      
    } else {
      
      case_set$DAM.Block.Energy.Profit <- case_set$DAM.Potential.Energy.Profit
      case_set$DAM.Block.Number <- case_set$DAM.Potential.Block.Number
      
    }
    
    case_set <- select(case_set, -DAM.Potential.Energy.Profit, -DAM.Potential.Block.Number)
    
    # 4. Sum Block-Level Profits and Costs -------------------
    
    case_set$DAM.All.Cost <- ifelse(!is.na(case_set$DAM.Block.Energy.Profit) & (is.na(lag(case_set$DAM.Block.Energy.Profit)) | case_set$DateTime == case_set$DateTime[1]),
                                    case_set$DAM.Cost + case_set$DAM.Startup.Cost,
                                    case_set$DAM.Cost)
    
    if (case_units_toggles$Toggles$Unit_Type == "CT") {
      case_set_block_level <- case_set %>% 
        filter(!is.na(Area) & !is.na(DAM.Block.Number)) %>%
        group_by(Area, DAM.Block.Number) %>% 
        summarise(DAM.Block.Hours = n(),
                  DAM.Block.Energy.Profit = sum(DAM.Block.Energy.Profit),
                  DAM.Block.Reserve.Profit = sum(DAM.Hourly.Reserve.Profit),
                  DAM.Block.LBMP = sum(LOE * DAM.LBMP),
                  RTD.Block.LBMP = sum(LOE * RT.LBMP),
                  RTD.Block.Reserve.Price = sum(RTD.Hourly.Reserve.Price),
                  DAM.Block.Costs = sum(DAM.All.Cost),
                  Block.Buyout.Fuel.Losses = sum(E.Buyout.Fuel.Loss.per.MWh) + max(E.Buyout.Startup.Fuel.Losses))
    }else{
      case_set_block_level <- case_set %>%
        filter(!is.na(Area) & !is.na(DAM.Block.Number)) %>%
        group_by(Area, DAM.Block.Number) %>%
        summarise(DAM.Block.Hours = n(),
                  DAM.Block.Energy.Profit = sum(DAM.Block.Energy.Profit),
                  DAM.Block.LBMP = sum(LOE * DAM.LBMP),
                  RTD.Block.LBMP = sum(LOE * RT.LBMP),
                  DAM.Block.Costs = sum(DAM.All.Cost),
                  Block.Buyout.Fuel.Losses = sum(E.Buyout.Fuel.Loss.per.MWh) + max(E.Buyout.Startup.Fuel.Losses))
    }
    # 5. Set the unit's commitments --------
    
    ### CTs:
    
    if(case_units_toggles$Toggles$Unit_Type == "CT") {
      case_set_block_commits <- case_set_block_level %>% mutate(DAM.Commit = ifelse(DAM.Block.Energy.Profit > 0 & DAM.Block.Energy.Profit > DAM.Block.Reserve.Profit,
                                                                                    "Energy",
                                                                                    NA))
      
      ### Decide whether the unit will stick to energy in real-time or buyout for reserves
      case_set_block_energy_commits <- filter(case_set_block_commits, DAM.Commit == "Energy") %>%
        mutate(RTD.Check.Energy = DAM.Block.LBMP - DAM.Block.Costs,
               RTD.Check.Buyout = DAM.Block.LBMP - RTD.Block.LBMP - Block.Buyout.Fuel.Losses,
               RTD.Check.Reserve = DAM.Block.LBMP - RTD.Block.LBMP - Block.Buyout.Fuel.Losses + RTD.Block.Reserve.Price)
      
      case_set_block_energy_commits <- as.data.table(case_set_block_energy_commits)
      case_set_block_energy_commits[RTD.Check.Energy > RTD.Check.Buyout & RTD.Check.Energy > RTD.Check.Reserve, ':=' (RTD.Dispatch = "Energy", 
                                                                                                                      RTD.Final.Hourly.Profit = RTD.Check.Energy / DAM.Block.Hours,
                                                                                                                      RTD.Final.Hourly.Profit.Other = 0,
                                                                                                                      RTD.Final.Hourly.Profit.Physical = RTD.Check.Energy / DAM.Block.Hours)]
      case_set_block_energy_commits[RTD.Check.Buyout >= RTD.Check.Energy & RTD.Check.Buyout >= RTD.Check.Reserve, ':=' (RTD.Dispatch = "Buyout",
                                                                                                                        RTD.Final.Hourly.Profit = RTD.Check.Buyout / DAM.Block.Hours,
                                                                                                                        RTD.Final.Hourly.Profit.Other = RTD.Check.Buyout / DAM.Block.Hours,
                                                                                                                        RTD.Final.Hourly.Profit.Physical = 0)]
      case_set_block_energy_commits[RTD.Check.Reserve > RTD.Check.Energy & RTD.Check.Reserve > RTD.Check.Buyout, ':=' (RTD.Dispatch = "Reserve",
                                                                                                                       RTD.Final.Hourly.Profit = RTD.Check.Reserve / DAM.Block.Hours,
                                                                                                                       RTD.Final.Hourly.Profit.Other = (DAM.Block.LBMP - RTD.Block.LBMP - Block.Buyout.Fuel.Losses) / DAM.Block.Hours,
                                                                                                                       RTD.Final.Hourly.Profit.Physical = RTD.Block.Reserve.Price / DAM.Block.Hours)]
    }else{
      ## CCs:
      case_set_block_commits <- case_set_block_level %>% mutate(DAM.Commit = ifelse(DAM.Block.Energy.Profit > 0, "Energy", NA))
      
      ### Decide whether the unit will stick to energy in real-time or buyout for reserves
      case_set_block_energy_commits <- filter(case_set_block_commits, DAM.Commit == "Energy") %>%
        mutate(RTD.Check.Energy = DAM.Block.LBMP - DAM.Block.Costs,
               RTD.Check.Buyout = DAM.Block.LBMP - RTD.Block.LBMP - Block.Buyout.Fuel.Losses)
      case_set_block_energy_commits <- as.data.table(case_set_block_energy_commits)
      case_set_block_energy_commits[RTD.Check.Energy > RTD.Check.Buyout, ':=' (RTD.Dispatch = "Energy", 
                                                                               RTD.Final.Hourly.Profit = RTD.Check.Energy / DAM.Block.Hours,
                                                                               RTD.Final.Hourly.Profit.Other = 0,
                                                                               RTD.Final.Hourly.Profit.Physical = RTD.Check.Energy / DAM.Block.Hours)]
      case_set_block_energy_commits[RTD.Check.Buyout >= RTD.Check.Energy, ':=' (RTD.Dispatch = "Buyout",
                                                                                RTD.Final.Hourly.Profit = RTD.Check.Buyout / DAM.Block.Hours,
                                                                                RTD.Final.Hourly.Profit.Other = RTD.Check.Buyout / DAM.Block.Hours,
                                                                                RTD.Final.Hourly.Profit.Physical = 0)]
    }
    
    # 6. Merge block-level commitments back onto unit commitments ---------------------------
    
    case_set_with_commits <- left_join(case_set, case_set_block_energy_commits, by = c("Area", "DAM.Block.Number")) %>%
      mutate(Fuel.Used = ifelse(RTD.Dispatch == "Energy", Economic.Fuel, "")) %>%
      select(-(DAM.Block.Hours:Block.Buyout.Fuel.Losses), -(contains("RTD.Check"))) %>%
      rename(DAM.Block.Energy.Profit = DAM.Block.Energy.Profit.x)
    
    # 7. Fill in remaining commitments -------------------------
    
    if(case_units_toggles$Toggles$Unit_Type == "CT") {
      case_set_with_commits$DAM.Commit <- case_when(is.na(case_set_with_commits$DAM.Commit) & case_set_with_commits$DAM.Hourly.Reserve.Profit > 0 ~ "Reserve", 
                                                    is.na(case_set_with_commits$DAM.Commit) ~ "None",
                                                    TRUE ~ case_set_with_commits$DAM.Commit)
    }else{
      case_set_with_commits$DAM.Commit <- ifelse(is.na(case_set_with_commits$DAM.Commit), "None", case_set_with_commits$DAM.Commit)
    }
    
    return(case_set_with_commits)
  }
  
  full_case_with_commits <- lapply(unit_list, dam_rt_commits)
  names(full_case_with_commits) <- unit_list
  set_dam_commits_output <- list(case_units_toggles, full_case_with_commits)
  names(set_dam_commits_output) <- c("Units_and_Toggles", "Full_case_Unit_Commits")
  return(set_dam_commits_output)
}

