###################################################################
# NYISO DCR Reset 2020: Net EAS Revenues Model
# Module 2 Part 3
#_________________________________________________________________

# Script Purpose
#   This script includes a set of nested functions that takes a case and determines real-time dispatch and buyout commitments for all units within that case
#   Inputs: a case number ("case_to_eval") and the full set of day-ahead commitments and real-time energy buyouts produced by the function in script 2.2 ("inputs")

set_rtd_commits <- function(inputs, case_to_eval) {
  case_base <- inputs[[paste0("Case_", case_to_eval)]]
  case_units_toggles <- case_base[[1]]
  case_inputs <- case_base[[2]]
  unit_list <- case_units_toggles$Unit_List
  unit_type <- case_units_toggles$Toggles$Unit_Type
  
  rtd_commits <- function(unit) {
    input_set <- case_inputs[[which(names(case_inputs) == unit)]]
    
    ## 1. Set Potential Real-Time Energy Profits ------------------------
    
    ### Establish the potential real-time energy profits where DAM Commit is not energy. 
    ### For instances where RTD Dispatch is already set to energy, it's set as the actual final hourly profit we already calculated. This is because
    ### we already evaluated profits for these units -- they were committed day-ahead for energy.
    ### Also add a potential startup flag for hours with positive profits that are preceded by hours with negative or nonzero profits
    
    if (unit_type == "CC") {
      input_set <- input_set %>% mutate(RTD.Potential.Energy.Profit = case_when(DAM.Commit == "None" ~ RTD.Hourly.Energy.Profit,
                                                                                RTD.Dispatch == "Energy" ~ RTD.Final.Hourly.Profit,
                                                                                TRUE ~ as.numeric(NA)),
                                        RTD.Potential.Startup = ifelse(RTD.Potential.Energy.Profit > 0 & lag(RTD.Potential.Energy.Profit) <= 0,
                                                                       "Startup",
                                                                       NA))
    }else{
      input_set <- input_set %>% mutate(RTD.Potential.Energy.Profit = case_when(DAM.Commit == "None" ~ RTD.Hourly.Energy.Profit,
                                                                                DAM.Commit == "Reserve" ~ (RTD.Hourly.Energy.Profit 
                                                                                                           - RTD.Hourly.Reserve.Price),
                                                                                RTD.Dispatch == "Energy" ~ RTD.Final.Hourly.Profit,
                                                                                TRUE ~ as.numeric(NA)),
                                        RTD.Potential.Startup = ifelse(RTD.Potential.Energy.Profit > 0 & lag(RTD.Potential.Energy.Profit) <= 0,
                                                                       "Startup",
                                                                       NA))
      ### Credit back Zones C and F for (1) buying fuel day-ahead instead of at the real-time price (reducing RT energy cost) and for (2) no longer having to pay a DA reserve cost, as that fuel cost now applies to energy provision (thus reducing DA reserve cost to 0)
      ### Set up a variable for updating RTD profits breakouts below as well -- credit (1) goes to RTD.Physical, credit (2) goes to RTD.Other
      ### Also, create a startup variable for hour where startup costs haven't yet been recovered, as startup costs in these hours should reflect DA fuel costs, not RT fuel costs. This will go to RTD.Physical
      input_set$RTD.Potential.Energy.Profit <- ifelse(input_set$Area %in% c("C", "F") & input_set$Economic.Fuel == "Gas" & input_set$DAM.Commit == "Reserve", 
                                                      input_set$RTD.Potential.Energy.Profit + 2 * (input_set$RTD.Gas.Cost - input_set$DAM.Gas.Cost),
                                                      input_set$RTD.Potential.Energy.Profit)
      input_set$DAM.Reserve.RTD.Energy.Addback <- ifelse(input_set$Area %in% c("C", "F") & input_set$Economic.Fuel == "Gas" & input_set$DAM.Commit == "Reserve",
                                                         input_set$RTD.Gas.Cost - input_set$DAM.Gas.Cost,
                                                         0)
      input_set$DAM.Reserve.RTD.Energy.Startup.Addback <- ifelse(input_set$Area %in% c("C", "F") & input_set$Economic.Fuel == "Gas" & input_set$DAM.Commit == "Reserve",
                                                                 input_set$RTD.Startup.Cost - input_set$DAM.Startup.Cost,
                                                                 0)
      
    }
    
    ### Set as NA all instances where RTD potential energy profits are less than zero
    input_set$RTD.Potential.Energy.Profit <- ifelse(input_set$RTD.Potential.Energy.Profit < 0, as.numeric(NA), input_set$RTD.Potential.Energy.Profit)
    
    ## 2. Flag for Energy Dispatch Based on Startup Cost Recovery Needs ----------------
    
    ### 2A. Add block numbers and hours in each block ---------------------------------
    
    #### Add block numbers to identify stretches where potential energy profits are greater than zero
    input_set_block_starts <- input_set %>% filter(!is.na(RTD.Potential.Energy.Profit) & is.na(lag(RTD.Potential.Energy.Profit))) %>%
      mutate(RTD.Potential.Block.Number = c(1:n()))
    
    input_set <- input_set %>% left_join(input_set_block_starts) %>%
      fill("RTD.Potential.Block.Number")
    
    input_set$RTD.Potential.Block.Number <- ifelse(is.na(input_set$RTD.Potential.Energy.Profit), as.numeric(NA), input_set$RTD.Potential.Block.Number)
    
    #### Add block hours for each observation to identify its place within each block (1st hour, 2nd hour, etc.)
    input_set_recovery_check <- input_set %>% filter(!is.na(RTD.Potential.Block.Number)) %>% group_by(Area, RTD.Potential.Block.Number) %>%
      mutate(RTD.Block.Hour = c(1:(n())))
    
    input_set_with_blocks <- input_set %>% left_join(input_set_recovery_check)
    
    ### 2B. Set dispatch behavior based on cost recovery windows -------------------------------
    
    #### For blocks that contain hours already set to dispatch energy in real-time, flag the entire block as having already incorporated startup costs.
    #### The flag here used for the block is RTD.Startup.Costs.Recovered; it's binary and is set to 1 if and only if an hour is part of a block [and hence
    #### has positive potential RTD profits] and there is an hour in the block with startup costs already incorporated.
    input_set_with_blocks$RTD.Energy.Dispatch.Flag <- case_when(is.na(input_set_with_blocks$RTD.Dispatch) ~ 0,
                                                                input_set_with_blocks$RTD.Dispatch == "Energy" ~ 1, 
                                                                TRUE ~ 0)
    input_set_startup_costs <- input_set_with_blocks %>% 
      group_by(Area, RTD.Potential.Block.Number) %>% 
      summarise(RTD.Startup.Costs.Recovered.Flag = max(RTD.Energy.Dispatch.Flag, na.rm = TRUE))
    input_set_with_blocks <- input_set_with_blocks %>% 
      left_join(input_set_startup_costs) %>% 
      select(-RTD.Energy.Dispatch.Flag)
    input_set_with_blocks[which(input_set_with_blocks$RTD.Startup.Costs.Recovered.Flag == 0),]$RTD.Startup.Costs.Recovered.Flag <- as.numeric(NA)
    
    #### For remaining potentially profitable blocks, calculate profits in each two-hour window
    input_set_profit_calc <- input_set_with_blocks %>% 
      mutate(RTD.2.Hour.Profit = case_when(!is.na(RTD.Potential.Energy.Profit) & is.na(RTD.Startup.Costs.Recovered.Flag) & is.na(lag(RTD.Potential.Energy.Profit)) ~ RTD.Potential.Energy.Profit,
                                           !is.na(RTD.Potential.Energy.Profit) & is.na(RTD.Startup.Costs.Recovered.Flag) ~ RTD.Potential.Energy.Profit + lag(RTD.Potential.Energy.Profit),
                                           TRUE ~ as.numeric(NA)))
    
    #### Set real-time behavior for aero units based on whether they can recover real-time startup costs within one hour.
    #### Set real-time behavior for F and J [CC & CT] units based on whether they can recover real-time startup costs within two hours.
    #### The Start.Hour.Flag identifies whether a given hour (aero units) or the hour preceding a given hour (F/J units) would be suitable for startup.
    #### This determination is based on whether profits in the 1-2 hour window in question exceed startup costs.
    #### If they do, the observation is assigned a flag value equal to the hour determined suitable for startup.
    
    if (unit %in% c("f", "j25", "j15", "ccj")) {
      input_set_amortize <- filter(input_set_profit_calc, !is.na(RTD.2.Hour.Profit) & !is.na(RTD.Startup.Cost))
      input_set_amortize$RTD.Start.Hour.Flag = case_when(input_set_amortize$DAM.Commit == "Reserve" & input_set_amortize$RTD.2.Hour.Profit > input_set_amortize$DAM.Startup.Cost ~ pmax(input_set_amortize$RTD.Block.Hour - 1, 1),
                                                         input_set_amortize$RTD.2.Hour.Profit > input_set_amortize$RTD.Startup.Cost ~ pmax(input_set_amortize$RTD.Block.Hour - 1, 1),
                                                         TRUE ~ as.numeric(NA))
      input_set_with_profits <- input_set_profit_calc %>% left_join(input_set_amortize)
    }else if(unit %in% c("aero")) {
      input_set_amortize <- filter(input_set_profit_calc, !is.na(RTD.Potential.Energy.Profit) & !is.na(RTD.Startup.Cost))
      input_set_amortize$RTD.Start.Hour.Flag = case_when((input_set_amortize$DAM.Commit == "Reserve" & (input_set_amortize$RTD.Potential.Energy.Profit > input_set_amortize$DAM.Startup.Cost) & is.na(input_set_amortize$RTD.Startup.Costs.Recovered.Flag)) ~ as.double(input_set_amortize$RTD.Block.Hour), 
                                                         input_set_amortize$RTD.Potential.Energy.Profit > input_set_amortize$RTD.Startup.Cost & is.na(input_set_amortize$RTD.Startup.Costs.Recovered.Flag) ~ as.double(input_set_amortize$RTD.Block.Hour),
                                                         TRUE ~ as.numeric(NA))
      input_set_with_profits <- input_set_profit_calc %>% left_join(input_set_amortize)
    }
    
    #### For each block, identify the earliest potential startup hour and set it as the block's startup hour.
    input_set_start_flags <- input_set_with_profits %>% filter(!is.na(RTD.Start.Hour.Flag)) %>%
      group_by(Area, RTD.Potential.Block.Number) %>%
      summarise(RTD.Block.Start.Hour = min(RTD.Start.Hour.Flag))
    input_set_with_profits <- input_set_with_profits %>% left_join(input_set_start_flags)
    
    ### 3. Set Dispatch Behavior and Profits for Units in Blocks with Potential Positive RTD Profits -------------
    #### Here, we set dispatch behavior and profits for units that have been identified in the block number process as having potential positive profit in real-time dispatch.
    input_set_dispatch_update <- copy(input_set_with_profits)
    setDT(input_set_dispatch_update)
    
    #### 3a. Set energy dispatch and fuel used for dispatch -------------------------------
    ##### For any hours that have startup costs already covered and have potential positive RTD profits but aren't dispatching yet,
    ##### OR for hours that have been identified already as having potential positive profits (in the block number process) 
    ##### and are either equal to or after the startup hour of the block they belong to, set them to dispatch energy in real time.
    input_set_dispatch_update[(RTD.Startup.Costs.Recovered.Flag == 1 & DAM.Commit != "Energy") | 
                                (!is.na(RTD.Block.Start.Hour) & RTD.Block.Start.Hour <= RTD.Block.Hour),
                              ':=' (RTD.Dispatch = "Energy",
                                    Fuel.Used = Economic.Fuel)]
    
    #### 3b. Set profits -------------------------------------------
    ##### Hours with DAM reserve commitments whose startup costs have already been covered, set profits as follows:
    ##### RTD.Final.Hourly.Profit = RTD energy profits (from dispatching in real-time) + DAM reserve profits (from offering reserves day-ahead)
    ##### RTD.Final.Hourly.Profit.Other = DAM reserve profits (from offering reserves day-ahead) - the price of securing RTD reserves (from not providing those reserves in real-time)
    ##### RTD.Final.Hourly.Profit.Physical = RTD energy profits (from dispatching in real-time) + a credit equivalent to the resale price of the fuel (from not having to resell the fuel)
    input_set_dispatch_update[RTD.Startup.Costs.Recovered.Flag == 1 & DAM.Commit == "Reserve",
                              ':=' (RTD.Final.Hourly.Profit = RTD.Potential.Energy.Profit + DAM.Hourly.Reserve.Profit,
                                    RTD.Final.Hourly.Profit.Other = DAM.Hourly.Reserve.Profit - RTD.Hourly.Reserve.Price + DAM.Reserve.RTD.Energy.Addback,
                                    RTD.Final.Hourly.Profit.Physical = RTD.Hourly.Energy.Profit + DAM.Reserve.RTD.Energy.Addback)]
    
    ##### For hours with no DAM commitments whose startup costs have already been covered, set profits as follows:
    ##### RTD.Final.Hourly.Profit = RTD energy profit
    ##### RTD.Final.Hourly.Profit.Other = 0
    ##### RTD.Final.Hourly.Profit.Physical = RTD energy profit [note: Potential.Energy.Profit and Hourly.Energy.Profit are the same here]
    input_set_dispatch_update[RTD.Startup.Costs.Recovered.Flag == 1 & DAM.Commit == "None", 
                              ':=' (RTD.Final.Hourly.Profit = RTD.Potential.Energy.Profit,
                                    RTD.Final.Hourly.Profit.Other = 0,
                                    RTD.Final.Hourly.Profit.Physical = RTD.Hourly.Energy.Profit)]
    
    ##### For hours with DAM reserve commitments whose startup costs have not already been covered and which are identified as startup hours for a block, set profits as follows:
    ##### RTD.Final.Hourly.Profit as RTD energy profits - RTD startup cost + DAM reserve profits
    ##### RTD.Final.Hourly.Profit.Other as DAM reserve profits - the price of securing RTD reserves
    ##### RTD.Final.Hourly.Profit.Physical as DAM reserve profits + fuel resale credit - RTD startup cost
    input_set_dispatch_update[!is.na(RTD.Block.Start.Hour) & RTD.Block.Start.Hour == RTD.Block.Hour & DAM.Commit == "Reserve", 
                              ':=' (RTD.Final.Hourly.Profit = RTD.Potential.Energy.Profit - RTD.Startup.Cost + DAM.Reserve.RTD.Energy.Startup.Addback + DAM.Hourly.Reserve.Profit,
                                    RTD.Final.Hourly.Profit.Other = DAM.Hourly.Reserve.Profit - RTD.Hourly.Reserve.Price + DAM.Reserve.RTD.Energy.Addback,
                                    RTD.Final.Hourly.Profit.Physical = RTD.Hourly.Energy.Profit + DAM.Reserve.RTD.Energy.Addback - RTD.Startup.Cost + DAM.Reserve.RTD.Energy.Startup.Addback)]
    
    ##### For hours with no DAM commitments whose startup costs have not already been covered and which are identified as startup hours for a block,
    ##### set profits as follows:
    ##### RTD.Final.Hourly.Profit = RTD energy profit  - RTD startup cost
    ##### RTD.Final.Hourly.Profit.Other = 0
    ##### RTD.Final.Hourly.Profit.Physical = RTD energy profit - RTD startup cost [note: Potential.Energy.Profit and Hourly.Energy.Profit are the same here]
    input_set_dispatch_update[!is.na(RTD.Block.Start.Hour) & RTD.Block.Start.Hour == RTD.Block.Hour & DAM.Commit == "None",
                              ':=' (RTD.Final.Hourly.Profit = RTD.Potential.Energy.Profit - RTD.Startup.Cost,
                                    RTD.Final.Hourly.Profit.Other = 0,
                                    RTD.Final.Hourly.Profit.Physical = RTD.Hourly.Energy.Profit - RTD.Startup.Cost)]
    
    ##### For hours with DAM reserve commitments whose startup costs have not already been covered and which are identified as RTD energy dispatch block hours that will 
    ##### be preceded by a startup hour, set profits as follows:
    ##### RTD.Final.Hourly.Profit = RTD energy profits (from dispatching in real-time) + DAM reserve profits (from offering reserves day-ahead)
    ##### RTD.Final.Hourly.Profit.Other = DAM reserve profits (from offering reserves day-ahead) - the price of securing RTD reserves (from not providing those reserves in real-time)
    ##### RTD.Final.Hourly.Profit.Physical = RTD energy profits (from offering reserves day-ahead) + a credit equivalent to the resale price of the fuel (from not having to resell the fuel)
    input_set_dispatch_update[!is.na(RTD.Block.Start.Hour) & RTD.Block.Start.Hour < RTD.Block.Hour & DAM.Commit == "Reserve",
                              ':=' (RTD.Final.Hourly.Profit = RTD.Potential.Energy.Profit + DAM.Hourly.Reserve.Profit,
                                    RTD.Final.Hourly.Profit.Other = DAM.Hourly.Reserve.Profit - RTD.Hourly.Reserve.Price + DAM.Reserve.RTD.Energy.Addback,
                                    RTD.Final.Hourly.Profit.Physical = RTD.Hourly.Energy.Profit + DAM.Reserve.RTD.Energy.Addback)]
    
    ##### For hours with no DAM commitments whose startup costs have not already been covered and which are identified as RTD energy dispatch block hours that will 
    ##### be preceded by a startup hour, set profits as follows:
    ##### RTD.Final.Hourly.Profit = RTD energy profit
    ##### RTD.Final.Hourly.Profit.Other = 0
    ##### RTD.Final.Hourly.Profit.Physical = RTD energy profit [note: Potential.Energy.Profit and Hourly.Energy.Profit are the same here]
    input_set_dispatch_update[!is.na(RTD.Block.Start.Hour) & RTD.Block.Start.Hour < RTD.Block.Hour & DAM.Commit == "None",
                              ':=' (RTD.Final.Hourly.Profit = RTD.Potential.Energy.Profit,
                                    RTD.Final.Hourly.Profit.Other = 0,
                                    RTD.Final.Hourly.Profit.Physical = RTD.Hourly.Energy.Profit)]
    
    
    
    input_set_dispatch_update <- as_tibble(input_set_dispatch_update) %>% select(-contains("Potential"), 
                                                                                 -RTD.Block.Hour,
                                                                                 -RTD.Start.Hour.Flag,
                                                                                 -RTD.Startup.Costs.Recovered.Flag, 
                                                                                 -RTD.2.Hour.Profit,
                                                                                 -RTD.Block.Start.Hour)
    
    ##### For CC units, we will need to reset profits in energy commits to replace $/Start MM costs with $/hr MM costs. 
    ##### (Note that RTD and DAM Startup MM Costs are identical, so we just use one variable.)
    ##### Here, we calculate the correction factor. We will apply after we set dispatch block numbers in Step 4 below.
    if(unit == "ccj"){
      input_set_dispatch_update <- mutate(input_set_dispatch_update, 
                                          MM.Cost.Adj = case_when(RTD.Dispatch == "Energy" & is.na(lag(RTD.Dispatch)) ~ DAM.Startup.MM.Cost,
                                                                  RTD.Dispatch == "Energy" ~ -1 * MM_COST / CAP,
                                                                  TRUE ~ as.numeric(NA)))
    }
    
    
    #### 4. Update Block Numbers -----------------------------
    ##### Here, we update the block numbers to reflect the changes to dispatch behavior after Section 3
    input_set_dispatch_blocks <- as_tibble(input_set_dispatch_update) %>% filter(RTD.Dispatch == "Energy" & (is.na(lag(RTD.Dispatch)) | 
                                                                                                               lag(RTD.Dispatch) == "Buyout" | 
                                                                                                               lag(RTD.Dispatch) == "Reserve" |
                                                                                                               DateTime == DateTime[1])) %>%
      group_by(Area) %>%
      mutate(RTD.Dispatch.Block.Number = c(1:n()))
    
    input_set_rtd_dispatch <- input_set_dispatch_update %>% left_join(input_set_dispatch_blocks) %>%
      fill("RTD.Dispatch.Block.Number")
    
    input_set_rtd_dispatch$RTD.Dispatch.Block.Number <- ifelse(input_set_rtd_dispatch$RTD.Dispatch != "Energy", NA, input_set_rtd_dispatch$RTD.Dispatch.Block.Number)
    
    ##### Apply correction factor for major maintenance costs for CCs
    if(unit == "ccj"){
      input_set_block_mm_adj <- group_by(input_set_rtd_dispatch, Area, RTD.Dispatch.Block.Number)
      input_set_block_mm_adj <- summarise(input_set_block_mm_adj,
                                          Block.Hours = n(),
                                          Block.MM.Cost.Adj = sum(MM.Cost.Adj, na.rm = TRUE))
      input_set_block_mm_adj <- mutate(input_set_block_mm_adj, Block.MM.Cost.Adj.Hrly = Block.MM.Cost.Adj / Block.Hours)
      
      input_set_rtd_dispatch <- left_join(input_set_rtd_dispatch, input_set_block_mm_adj)
      input_set_rtd_dispatch <- mutate(input_set_rtd_dispatch,
                                       RTD.Final.Hourly.Profit = ifelse(is.na(RTD.Final.Hourly.Profit), RTD.Final.Hourly.Profit, RTD.Final.Hourly.Profit + Block.MM.Cost.Adj.Hrly),
                                       RTD.Final.Hourly.Profit.Physical = ifelse(is.na(RTD.Final.Hourly.Profit.Physical),
                                                                                 RTD.Final.Hourly.Profit.Physical,
                                                                                 RTD.Final.Hourly.Profit.Physical + Block.MM.Cost.Adj.Hrly))
      input_set_rtd_dispatch <- select(input_set_rtd_dispatch, -MM.Cost.Adj, -Block.MM.Cost.Adj, -Block.Hours)
    }
    
    
    #### 5. Review Non-Energy Hours for Possible Switching between Reserve & None ----------------
    ##### Here, we look at hours set to dispatch reserves or nothing in real time in order to check whether or not they
    ##### switch their commitments -- to buy out of their reserve commitment or to commit to dispatching reseves.
    
    ##### First -- CC units do not provide reserves, so we set all their non-energy dispatch actions and profits to none/0.
    if (unit_type == "CC") {
      colT <- colnames(input_set_rtd_dispatch)
      colnames(input_set_rtd_dispatch) = paste0("col", 1:ncol(input_set_rtd_dispatch))
      setDT(input_set_rtd_dispatch)
      colnames(input_set_rtd_dispatch) = colT
      input_set_rtd_dispatch[is.na(RTD.Dispatch),
                             ':=' (RTD.Dispatch = "None",
                                   RTD.Final.Hourly.Profit = 0,
                                   RTD.Final.Hourly.Profit.Other = 0,
                                   RTD.Final.Hourly.Profit.Physical = 0)]
    }else{
      colT <- colnames(input_set_rtd_dispatch)
      colnames(input_set_rtd_dispatch) = paste0("col", 1:ncol(input_set_rtd_dispatch))
      setDT(input_set_rtd_dispatch)
      colnames(input_set_rtd_dispatch) = colT
      
      ##### For units not committed to providing reserves day-ahead, examine whether it's profitable to provide them.
      ##### If so, profits equal revenues from selling the RTD reserve product minus the fuel resale cost
      input_set_rtd_dispatch[DAM.Commit == "None" & (RTD.Hourly.Reserve.Price - E.Buyout.Fuel.Loss.per.MWh) > 0 & (RTD.Dispatch != "Energy" | is.na(RTD.Dispatch)),
                             ':=' (RTD.Dispatch = "Reserve",
                                   RTD.Final.Hourly.Profit = RTD.Hourly.Reserve.Price - E.Buyout.Fuel.Loss.per.MWh,
                                   RTD.Final.Hourly.Profit.Other = 0,
                                   RTD.Final.Hourly.Profit.Physical = RTD.Hourly.Reserve.Price - E.Buyout.Fuel.Loss.per.MWh)]
      ##### Otherwise, their behavior doesn't change -- they do not dispatch and receive no profits.
      input_set_rtd_dispatch[DAM.Commit == "None" & (RTD.Hourly.Reserve.Price - E.Buyout.Fuel.Loss.per.MWh) <= 0 & (RTD.Dispatch != "Energy" | is.na(RTD.Dispatch)),
                             ':=' (RTD.Dispatch = "None",
                                   RTD.Final.Hourly.Profit = 0,
                                   RTD.Final.Hourly.Profit.Other = 0,
                                   RTD.Final.Hourly.Profit.Physical = 0)]
      
      ##### For units committed to providing reserves day-ahead, if the real-time price of reserve provision is zero or negative, buy out of reserves.
      input_set_rtd_dispatch[DAM.Commit == "Reserve" & RTD.Hourly.Reserve.Price <= 0 & (RTD.Dispatch != "Energy" | is.na(RTD.Dispatch)),
                             ':=' (RTD.Dispatch = "Buyout",
                                   RTD.Final.Hourly.Profit = DAM.Hourly.Reserve.Profit - RTD.Hourly.Reserve.Price,
                                   RTD.Final.Hourly.Profit.Other = DAM.Hourly.Reserve.Profit - RTD.Hourly.Reserve.Price,
                                   RTD.Final.Hourly.Profit.Physical = 0)]
      ##### Otherwise, their behavior doesn't change -- they stick with their DAM commitment and are compensated accordingly.
      input_set_rtd_dispatch[DAM.Commit == "Reserve" & RTD.Hourly.Reserve.Price > 0 & (RTD.Dispatch != "Energy" | is.na(RTD.Dispatch)),
                             ':=' (RTD.Dispatch = "Reserve",
                                   RTD.Final.Hourly.Profit = DAM.Hourly.Reserve.Profit,
                                   RTD.Final.Hourly.Profit.Other = 0,
                                   RTD.Final.Hourly.Profit.Physical = DAM.Hourly.Reserve.Profit)]
      input_set_rtd_dispatch <- as.data.frame(input_set_rtd_dispatch)
    }
    return(input_set_rtd_dispatch)
  }

 
  full_case_with_commits <- lapply(unit_list, rtd_commits)
  
  names(full_case_with_commits) <- unit_list
  set_dam_commits_output <- list(case_units_toggles, full_case_with_commits)
  names(set_dam_commits_output) <- c("Units_and_Toggles", "Full_Case_Unit_Commits")
  return(set_dam_commits_output)
}
