###################################################################
# NYISO DCR Reset 2020: Net EAS Revenues Model
# Module 2, Part 4
#_________________________________________________________________

# Script Purpose
#   This script includes a set of nested functions that take a case and adjusts unit profits and operations for forced outages and environmental runtime limitations.
#   Inputs: a case number ("case_to_eval") and the full set of real-time dispatch decisions produced by the function in script 2.3 ("inputs")

# 0. Master Function -------------------------------------

eford_plus_runtime_adj <- function(inputs, case_to_eval) {
  case_base <<- inputs[[paste0("Case_", case_to_eval)]]
  case_units_toggles <<- case_base[[1]]
  case_inputs <<- case_base[[2]]
  unit_list <<- case_units_toggles$Unit_List
  scr <<- tolower(case_units_toggles$Toggles$SCR_Toggle)
  if (case_units_toggles$Toggles$Unit_Type == "CT") {
    case_inputs <- lapply(unit_list, runtime_limits, inputs = case_inputs, scr_toggle = scr) %>% set_names(unit_list)
    units_eford_runtime_output <- lapply(unit_list, EFORd_adj, inputs = case_inputs)
  } else if (case_units_toggles$Toggles$Unit_Type == "CC") {
    case_inputs <- lapply(unit_list, cc_min_runtimes, inputs = case_inputs) %>% set_names(unit_list)
    units_eford_runtime_output <- lapply(unit_list, EFORd_adj, inputs = case_inputs)
  } else {
    print("Error: unit type toggle incorrectly specified")
    break
  }
  names(units_eford_runtime_output) <- unit_list
  units_adj_and_toggles <- list(case_units_toggles, units_eford_runtime_output)
  names(units_adj_and_toggles) <- c("Units_and_Toggles", "Final_Hourly_Results")
  return(units_adj_and_toggles)
}

# 1. EFORd Adjustment Function ------------------------------
## Dock profits according to EFORd outage likelihood

EFORd_adj <- function(inputs, unit) {
  unit_inputs <- inputs[[unit]] %>% mutate(EFORd.Profit.Loss = RTD.Final.Hourly.Profit.Physical * (1 - EFORd))
  unit_inputs$RTD.Final.Hourly.Profit <- unit_inputs$RTD.Final.Hourly.Profit - unit_inputs$EFORd.Profit.Loss
  return(unit_inputs)
}

# 2. NSPS Runtime Limitations Function ----------------------

runtime_limits <- function(inputs, unit, scr_toggle) {
  unit_inputs <- inputs[[unit]]
  scr_setting <- ifelse(scr_toggle == "yes", "scr", "noscr")
  print(paste0("Applying NSPS limit to ", unit))
  techparams_input_loc <- case_units_toggles$Toggles$Technical_Inputs_Filename
  source_wb <- ifelse(techparams_input_loc == "[Default]",
                      paste0(shared_inputs_folder, "/Unit Technical Parameters.xlsx"),
                      paste0(run_inputs_folder, techparams_input_loc))
  if(scr_setting == "scr"){
    all_runtime_limits <- read_xlsx(source_wb, sheet = "CTs_runtimes_scr")  
    unit_runtime_limit <- unlist(all_runtime_limits[1,which(names(all_runtime_limits) == paste0("rt_limit_", unit, "_",  tolower(scr_setting)))])/8760
    
    ## If there's no limit, immediately return the inputs
    if (unit_runtime_limit == 1){
      print("No runtime limit")
      return(unit_inputs)
    }
    
    ## 2A. Assess Whether Runtime Limits are Needed and, if so, How Much ---------------------
    
    ### Calculate the key runtime parameters: energy dispatch hours, limit, and overage
    runtime_table <<- unit_inputs %>% 
      group_by(Model.Year, Area) %>%
      mutate(RTD.Energy.Flag = ifelse(RTD.Dispatch == "Energy", 1, 0)) %>%
      summarise(Model.Year.Run.Hours = sum(RTD.Energy.Flag, na.rm = TRUE),
                Model.Year.Total.Hours = n()) %>%
      mutate(Model.Year.Run.Limit = unit_runtime_limit * Model.Year.Total.Hours,
             Model.Year.Overage = pmax(0, Model.Year.Run.Hours - Model.Year.Run.Limit)) %>%
      filter(Model.Year.Overage > 0)
    
    ### Merge limits with the original input set, filtered for hours with RTD energy dispatch in the areas to which runtime limits are applied
    unit_inputs_w_runtimes <<- unit_inputs %>% filter(RTD.Dispatch == "Energy") %>%
      select(Area, Zone, Model.Year, RTD.Dispatch.Block.Number, DAM.Hourly.Reserve.Profit, RTD.Final.Hourly.Profit, EFORd) %>%
      full_join(runtime_table) %>%
      filter(!is.na(Model.Year.Overage))
    
    if(nrow(unit_inputs_w_runtimes) == 0) {
      print("No years exceeding the overage limit")
      return(unit_inputs)
    }
    
    ## 2B. Decide Which Hours to Turn Off to Meet Runtime Limits --------------
    
    ### Calculate lost average hourly profits for each block subjected to runtime limits. Order the resulting data frame by lost profits (smallest to largest).
    reserve_switch_lost_profits <<- unit_inputs_w_runtimes %>% 
      group_by(Model.Year, Zone, Area, RTD.Dispatch.Block.Number, Model.Year.Overage) %>%
      summarise(Block.Hours = n(),
                Block.Reserve.Profits = sum(pmax(0, DAM.Hourly.Reserve.Profit * EFORd)),
                Block.Hrly.Reserve.Profits = Block.Reserve.Profits / Block.Hours,
                Block.Run.Profits = sum(RTD.Final.Hourly.Profit),
                Block.Hrly.Run.Profits = Block.Run.Profits / Block.Hours) %>%
      mutate(Block.Foregone.Profit = Block.Run.Profits - Block.Reserve.Profits,
             Block.Foregone.Hrly.Profit = Block.Hrly.Run.Profits - Block.Hrly.Reserve.Profits) %>%
      arrange(Model.Year, Zone, Area, Block.Foregone.Hrly.Profit) %>%
      ungroup()
    
    ### Starting from the block with the lowest profits per hour, flag blocks to be turned off until the overage limit is reached
    turn_off_flags <- function(yr, area) {
      reserve_switch_year <- filter(reserve_switch_lost_profits, Model.Year == yr, Area == area) %>%
        mutate(Cum.Block.Hours = cumsum(Block.Hours),
               Turn.Off.Flag = ifelse(lag(Cum.Block.Hours) < Model.Year.Overage, 1, 0))
      #### Replace NA in Turn.Off.Flag's first observation with a 1, since the first observation doesn't have a lag variable
      reserve_switch_year$Turn.Off.Flag[1] <- 1
      
      #### The final block flagged for turning off may be a long block that ends up censoring an unnecessary amount of profits. It may be more profitable to turn off the remaining block(s) needed based purely on 
      #### lowest total profits (rather than profits per hr); check this. If so, reassign.
      last_hour <- filter(reserve_switch_year, Turn.Off.Flag == 1 & lead(Turn.Off.Flag) == 0)
      overage_remaining <- last_hour$Block.Hours - (last_hour$Cum.Block.Hours - last_hour$Model.Year.Overage)
      if (overage_remaining > 0) {
        remaining_hours <- filter(reserve_switch_year, Turn.Off.Flag == 0) %>%
          arrange(Block.Foregone.Profit) %>%
          mutate(Cuml.Hours = cumsum(Block.Hours),
                 Cuml.Foregone.Profit = cumsum(Block.Foregone.Profit),
                 overage = overage_remaining) %>%
          filter(is.na(lag(Cuml.Hours)) | lag(Cuml.Hours) < overage)
        alternative_profits_lost <- max(remaining_hours$Cuml.Foregone.Profit)
        if(alternative_profits_lost < last_hour$Block.Foregone.Profit) {
          reserve_switch_year[which(reserve_switch_year$RTD.Dispatch.Block.Number %in% remaining_hours$RTD.Dispatch.Block.Number), ]$Turn.Off.Flag <- 1
          reserve_switch_year[which(reserve_switch_year$RTD.Dispatch.Block.Number == last_hour$RTD.Dispatch.Block.Number), ]$Turn.Off.Flag <- 0
        }        
      }
      
      
      return(reserve_switch_year)
    }
    
    reserve_switch_full <- map2(runtime_table$Model.Year, runtime_table$Area, turn_off_flags) %>% rbindlist()
    
    
  }else{
    
    ## Add hourly NOx emissions volumes
    inputs_w_emissions <- unit_inputs %>% filter(RTD.Dispatch == "Energy") %>% mutate(NOx.Emissions = ifelse(Fuel.Used == "Gas", CAP * NOX_NG * 2000, CAP * NOX_OIL * 2000))
    nox_limits <- read_xlsx(source_wb, sheet = "nox_limits")
    
    ## 2A. Assess Whether Runtime Limits are Needed and, if so, How Much ---------------------
    ### Calculate the key runtime parameters: energy dispatch hours, limit, and overage
    runtime_table <- inputs_w_emissions %>%
      group_by(Model.Year, Area) %>%
      summarise(Total.NOx.Emissions = sum(NOx.Emissions, na.rm = TRUE)) %>%
      ungroup() %>%
      left_join(nox_limits)%>%
      mutate(Model.Year.Overage = pmax(0, Total.NOx.Emissions - NOx.Limit)) %>%
      filter(Model.Year.Overage > 0)
    
    if(nrow(runtime_table) == 0){
      print("No years exceeding the overage limit")
      return(unit_inputs)
    }
    
    ### Merge limits with the original input set, filtered for hours with RTD energy dispatch
    unit_inputs_w_runtimes <<- inputs_w_emissions %>% filter(RTD.Dispatch == "Energy") %>%
      select(Area, Zone, Model.Year, RTD.Dispatch.Block.Number, DAM.Hourly.Reserve.Profit, RTD.Final.Hourly.Profit, Fuel.Used, NOx.Emissions, EFORd) %>%
      full_join(runtime_table) %>%
      filter(!is.na(Model.Year.Overage)) %>%
      select(-Model.Year.Overage)
    
    ## 2B. Decide Which Hours to Turn Off to Meet Runtime Limits --------------
    switch_lost_profits <- unit_inputs_w_runtimes %>% 
      group_by(Model.Year, Zone, Area, RTD.Dispatch.Block.Number) %>%
      summarise(Block.NOx.Emissions = sum(NOx.Emissions),
                Block.Reserve.Profits = sum(pmax(0, DAM.Hourly.Reserve.Profit * EFORd)),
                Block.Reserve.Profits.per.NOx = Block.Reserve.Profits / Block.NOx.Emissions,
                Block.Run.Profits = sum(RTD.Final.Hourly.Profit),
                Block.Run.Profits.per.NOx = Block.Run.Profits / Block.NOx.Emissions,
                Block.Hours = n()) %>%
      mutate(Block.Foregone.Profit.per.NOx = Block.Run.Profits.per.NOx - Block.Reserve.Profits.per.NOx,
             Block.Foregone.Profit = Block.Run.Profits - Block.Reserve.Profits) %>%
      arrange(Model.Year, Zone, Area, Block.Foregone.Profit.per.NOx) %>%
      ungroup()
    
    ### Starting from the block with the lowest profits per lb of NOx emissions, flag blocks to be turned off until the overage limit is reached. Do so one area & model year at a time.
    turn_off_flags_noscr <- function(yr, area) {
      switch_hours <- filter(switch_lost_profits, Model.Year == yr, Area == area) %>% mutate(Turn.Off.Flag = 0)
      overage <- filter(runtime_table, Model.Year == yr, Area == area)$Model.Year.Overage
      counter <- 1
      while(overage > 0){
        switch_hours[counter, "Turn.Off.Flag"] <- 1
        overage = overage - switch_hours[counter,]$Block.NOx.Emissions
        counter <- counter + 1
      }
      full_switch_hours <- switch_hours
      #### The final block flagged for turning off may be a large block that ends up censoring an unnecessary amount of profits It may be more profitable to turn off remaining block(s) needed based
      #### upon lowest total profits (rather than profits per NOx lb emitted); check this. If so, reassign.
      last_hour <- filter(full_switch_hours, Turn.Off.Flag == 1 & lead(Turn.Off.Flag) == 0)
      overage_remaining <- last_hour$Block.NOx.Emissions + overage
      remaining_hours <- filter(full_switch_hours, Turn.Off.Flag == 0) %>%
        arrange(Block.Foregone.Profit) %>%
        mutate(Cuml.Emissions = cumsum(Block.NOx.Emissions),
               Cuml.Foregone.Profit = cumsum(Block.Foregone.Profit),
               overage = overage_remaining) %>%
        filter(is.na(lag(Cuml.Emissions)) | lag(Cuml.Emissions) < overage)
      alternative_profits_lost <- max(remaining_hours$Cuml.Foregone.Profit)
      if(alternative_profits_lost < last_hour$Block.Foregone.Profit) {
        full_switch_hours[which(full_switch_hours$RTD.Dispatch.Block.Number %in% remaining_hours$RTD.Dispatch.Block.Number), ]$Turn.Off.Flag <- 1
        full_switch_hours[which(full_switch_hours$RTD.Dispatch.Block.Number == last_hour$RTD.Dispatch.Block.Number), ]$Turn.Off.Flag <- 0
      }
      
      return(full_switch_hours)
    }
    
    reserve_switch_full <- map2(runtime_table$Model.Year, runtime_table$Area, turn_off_flags_noscr) %>% rbindlist()
  } 
  
  ### Merge turn off flags onto hour-level observations. Hours selected to be turned off are allowed to provide reserves, if reserves are profitable.
  reserve_switch_for_merge <<- select(reserve_switch_full, Model.Year, Area, RTD.Dispatch.Block.Number, Turn.Off.Flag)
  unit_inputs_limited <<- full_join(unit_inputs, reserve_switch_for_merge, by = c("Model.Year", "Area", "RTD.Dispatch.Block.Number"))
  unit_inputs_limited <- as.data.table(unit_inputs_limited)
  unit_inputs_limited[Turn.Off.Flag == 1, ':=' (RTD.Final.Hourly.Profit = pmax(DAM.Hourly.Reserve.Profit, 0),
                                                RTD.Final.Hourly.Profit.Physical = pmax(DAM.Hourly.Reserve.Profit, 0),
                                                RTD.Final.Hourly.Profit.Other = 0,
                                                RTD.Dispatch = "Limited")]
  
  return(unit_inputs_limited)
}

# 3. Minimum Runtimes Function -----------------------------
## This function evaluates whether CC units would save money in between energy dispatch blocks by staying online, so as to avoid incurring a second round of startup costs.

cc_min_runtimes <- function(inputs, unit) {
  ## Generate a flag for each block dispatched in real-time that indicates whether the subsequent starting block is subject to DAM startup costs or to RTD startup costs.
  unit_inputs <- inputs[[unit]]
  dam_or_rtd_blocks <- unit_inputs %>% 
    filter(!is.na(RTD.Dispatch.Block.Number)) %>%
    select(Area, RTD.Dispatch.Block.Number, DAM.Commit, RTD.Dispatch) %>%
    group_by(Area, RTD.Dispatch.Block.Number) %>%
    mutate(DAM.Energy.Flag = ifelse(DAM.Commit == "Energy", 1, 0)) %>%
    summarise(DAM.Startup.Block = max(DAM.Energy.Flag)) %>%
    mutate(Next.Startup.DAM = lead(DAM.Startup.Block)) %>%
    rename(Block.Number = RTD.Dispatch.Block.Number) %>%
    select(-DAM.Startup.Block)
  
  ## Set up inter-blocks between energy dispatch hours
  unit_inputs$Inter.Block.Flag <- ifelse(is.na(unit_inputs$RTD.Dispatch.Block.Number), 1, NA)
  unit_inputs_inter_block_heads <- unit_inputs %>% 
    filter(!is.na(Inter.Block.Flag) & is.na(lag(Inter.Block.Flag)) & DateTime != ymd_hms("2013-09-01 00:00:00") & DateTime != ymd_hms("2016-08-31 23:00:00")) %>%
    select(DateTime, Area, Inter.Block.Flag) %>%
    group_by(Area) %>%
    mutate(Block.Number = c(1:n())) %>%
    select(-Inter.Block.Flag)
  
  ## Merge the block numbers onto the real-time block dispatch dataset with DAM/RTD startup flags. This creates a dataset of inter-block numbers
  ## with flags for whether the startup costs to compare for each inter-block are RTD or DAM.
  unit_inputs_set_startup_flags <- full_join(dam_or_rtd_blocks, unit_inputs_inter_block_heads) %>%
    rename(Inter.Block.Number = Block.Number)
  
  ## Merge the inter-block numbers with startup flags onto the original input set and fill the inter-block numbers down
  ## Do this area-by-area so that fill commands don't extend across area lines. 
  ## Reassign inter-block numbers after filling to set all periods after the final startup of the model to NA.
  apply_interblocknums <- function(area) {
    area_inputs <- filter(unit_inputs, Area == area)
    area_inputs_inter_blocks <- area_inputs %>% 
      left_join(unit_inputs_set_startup_flags) %>%
      fill(Inter.Block.Number) %>%
      setDF()
    area_inputs_inter_blocks$Inter.Block.Number <- ifelse(is.na(area_inputs_inter_blocks$Inter.Block.Flag) | 
                                                            area_inputs_inter_blocks$DateTime > 
                                                            dplyr::last(filter(area_inputs_inter_blocks, RTD.Dispatch == "Energy")$DateTime), 
                                                          NA,
                                                          area_inputs_inter_blocks$Inter.Block.Number)
    return(area_inputs_inter_blocks)
  }
  
  area_vec <- unique(unit_inputs$Area)
  unit_inputs_inter_blocks <- lapply(area_vec, apply_interblocknums) %>%
    set_names(area_vec) %>% bind_rows()
  
  ## Add next-hour startup costs
  unit_inputs_inter_blocks <- mutate(unit_inputs_inter_blocks,
                                     Next.DAM.Startup.Cost = lead(DAM.Startup.Cost),
                                     Next.RTD.Startup.Cost = lead(RTD.Startup.Cost))
  
  ## Calculate block-level profits and savings from starting up during a given inter-block at minimum load (decisions made on a $/start MM cost basis,
  ##  costs allocated on a $/hr basis)
  inter_blocks <- unit_inputs_inter_blocks %>%
    filter(!is.na(Inter.Block.Number)) %>%
    group_by(Area, Inter.Block.Number) %>%
    summarise(Compare.DAM.Startup = first(Next.Startup.DAM),
              Inter.Block.Hours = n(),
              Inter.Block.Energy.Profit.OpBasis = sum(RTD.Hourly.Energy.Min.Profit, na.rm = TRUE),
              DAM.Startup.Cost.OpBasis = last(Next.DAM.Startup.Cost, na.rm = TRUE),
              RTD.Startup.Cost.OpBasis = last(Next.RTD.Startup.Cost, na.rm = TRUE),
              DAM.Startup.Loss = min(E.Buyout.Startup.Fuel.Losses, na.rm = TRUE),
              # Note: MM_COST is constant for all hours, and CAP, DAM.Startup.MM.Cost, and RTD.Startup.MM.Cost are constant for all hours in a given area
              # Taking the mean is just a way of keeping these variables
              MM_COST = mean(MM_COST),
              CAP = mean(CAP),
              DAM.Startup.MM.Cost = mean(DAM.Startup.MM.Cost),
              RTD.Startup.MM.Cost = mean(RTD.Startup.MM.Cost)) %>%
    mutate(Saved.Startup.Expense.OpBasis = ifelse(Compare.DAM.Startup == 1, 
                                          Inter.Block.Energy.Profit.OpBasis + DAM.Startup.Cost.OpBasis - DAM.Startup.Loss,
                                          Inter.Block.Energy.Profit.OpBasis + RTD.Startup.Cost.OpBasis),
           Saved.Startup.Expense = ifelse(Compare.DAM.Startup == 1, 
                                          Inter.Block.Energy.Profit.OpBasis - Inter.Block.Hours * MM_COST / CAP + DAM.Startup.Cost.OpBasis - DAM.Startup.MM.Cost - DAM.Startup.Loss,
                                          Inter.Block.Energy.Profit.OpBasis - Inter.Block.Hours * MM_COST / CAP + RTD.Startup.Cost.OpBasis - RTD.Startup.MM.Cost)) %>%
    filter(Inter.Block.Hours <= 24, Saved.Startup.Expense.OpBasis > 0) %>%
    select(-(Inter.Block.Energy.Profit.OpBasis:DAM.Startup.Loss))
  
  unit_inputs_inter_dispatch <- left_join(unit_inputs_inter_blocks, inter_blocks, by = c("Area", "Inter.Block.Number")) %>%
    mutate(Inter.Block.Offset = Saved.Startup.Expense / Inter.Block.Hours) %>%
    rename(MM_COST = MM_COST.x, CAP = CAP.x, DAM.Startup.MM.Cost = DAM.Startup.MM.Cost.x, RTD.Startup.MM.Cost = RTD.Startup.MM.Cost.x) %>%
    select(-Inter.Block.Flag, -(Next.Startup.DAM:Saved.Startup.Expense))
  
  colT <- colnames(unit_inputs_inter_dispatch)
  colnames(unit_inputs_inter_dispatch) = paste0("col", 1:ncol(unit_inputs_inter_dispatch))
  setDT(unit_inputs_inter_dispatch)
  colnames(unit_inputs_inter_dispatch) = colT
  
  ## Reset all observations where saved expenses from continued operation between energy dispatch blocks are positive. Now, these observations are set to dispatch and their revenues reflect
  ## the hourly savings from dispatching. This way, there's no need to undo the application of startup costs to the hours where they had already been applied -- this is 
  ## just shifting the same total revenue across different hours.
  
  
  unit_inputs_inter_dispatch[!is.na(Inter.Block.Offset), ':=' (RTD.Dispatch = "Energy",
                                                               RTD.Final.Hourly.Profit = Inter.Block.Offset,
                                                               RTD.Final.Hourly.Profit.Other = 0,
                                                               RTD.Final.Hourly.Profit.Physical = Inter.Block.Offset,
                                                               Fuel.Used = Economic.Fuel)]
  return(unit_inputs_inter_dispatch)
}
