##################################################################
# NYISO DCR Reset 2020  
# Module 3, Part 1                                                
#_________________________________________________________________

# Script Purpose
#   This script includes a set of nested functions that take a case and summarize behavior of all units within the case in a series of tables.
#   Inputs: a case number ("case_to_eval") and the full results of the model logic produced by the functions in script 2.4 ("inputs")

# 1. Overall Function To Be Executed in the Control --------------------------------------------
## This function summarizes the results of the model in a series of tables to be exported to Excel.

all_unit_summaries <- function(inputs, case_to_eval) {
  case_base <- inputs[[paste0("Case_", case_to_eval)]]
  case_units_toggles <- case_base[[1]]
  hrly_results <- case_base[[2]]
  scr_setting <- case_units_toggles$Toggles$SCR_Toggle
  unit_type <- case_units_toggles$Toggles$Unit_Type
  unit_list <- case_units_toggles$Unit_List
  results <- lapply(unit_list, summary_tbls, 
                    case_num = case_to_eval, 
                    full_hrly_results = hrly_results, 
                    unit_type = unit_type) %>% 
    set_names(unit_list) 
  return(results)
}

# 2. Unit Summary Function ----------------------------------------------------
## This code generates summary tables for a single specified unit.

summary_tbls <- function(full_hrly_results, unit, case_num, unit_type) {
  print(paste0("Evaluating case no. ", case_num, ", unit ", unit))
  hrly_results <- full_hrly_results[[unit]]
  
  ## 2A. Set up complete dispatch hourly results to produce summary tables by adding countable flags for categorical variables. ------
  
  ### Add flags to be used for summation of dispatch behavior.
  
  hrly_results = hrly_results %>% mutate(RTD.Dispatch=case_when(
    RTD.Dispatch=="Discharge" ~ "Energy",
    TRUE ~ RTD.Dispatch
  ))
  
  dispatch_sumbase <<- hrly_results %>%
    mutate(Unit.Start = case_when(is.na(RTD.Dispatch) ~ 0,
                                  RTD.Dispatch == "Energy" & 
                                    (lag(RTD.Dispatch) != "Energy" | DateTime == ymd_hms("2013-09-01 00:00:00"))  ~ 1,
                                  TRUE ~ as.numeric(NA)),
           Month = month(DateTime),
           Energy.Flag = case_when(is.na(RTD.Dispatch) ~ 0,
                                   RTD.Dispatch == "Energy" ~ 1,
                                   TRUE ~ as.numeric(NA)),
           Reserve.Flag = case_when(is.na(RTD.Dispatch) ~ 0,
                                    RTD.Dispatch == "Reserve" ~ 1,
                                    TRUE ~ as.numeric(NA)),
           Buyout.Flag = case_when(is.na(RTD.Dispatch) ~ 0,
                                   RTD.Dispatch == "Buyout" ~ 1,
                                   TRUE ~ as.numeric(NA)),
           Limited.Flag = case_when(is.na(RTD.Dispatch) ~ 0,
                                    RTD.Dispatch == "Limited" ~ 1,
                                    TRUE ~ as.numeric(NA)),
           None.Flag = case_when(is.na(RTD.Dispatch) ~ 0,
                                 RTD.Dispatch == "None" ~ 1,
                                 TRUE ~ as.numeric(NA)),
           Energy.Profit = case_when(is.na(RTD.Dispatch) ~ 0,
                                     RTD.Dispatch == "Energy" ~ RTD.Final.Hourly.Profit,
                                     TRUE ~ as.numeric(NA)),
           Reserve.Profit = case_when(is.na(RTD.Dispatch) ~ 0,
                                      RTD.Dispatch == "Reserve" ~ RTD.Final.Hourly.Profit,
                                      TRUE ~ as.numeric(NA)),
           Buyout.Profit = case_when(is.na(RTD.Dispatch) ~ 0,
                                     RTD.Dispatch == "Buyout" ~ RTD.Final.Hourly.Profit,
                                     TRUE ~ as.numeric(NA)),
           Limited.Profit = case_when(is.na(RTD.Dispatch) ~ 0,
                                      RTD.Dispatch == "Limited" ~ RTD.Final.Hourly.Profit,
                                      TRUE ~ as.numeric(NA)),
           None.Profit = case_when(is.na(RTD.Dispatch) ~ 0,
                                   RTD.Dispatch == "None" ~ RTD.Final.Hourly.Profit,
                                   TRUE ~ as.numeric(NA)))
  dispatch_sumbase <<- dispatch_sumbase %>% mutate(Fuel.Used = ifelse(is.na(Fuel.Used) | Fuel.Used == "", "None", Fuel.Used))
  
  
  ## 2B. Generate summary tables -----------------------------------------------------------
  
  ### Set base grouping vector for summary tables. All summary tables will have results that are at least grouped by these variables
  
  group_vec <- c("Area", "Zone", "Capacity_Zone")
  
  ### Generate full diagnostic summary table set by calling the function in Section 3.
  
  summary_tables <- gen_summary_tables(dispatch_sumbase, unit_type, group_vec = group_vec)
  return(summary_tables)
  
}

# 3. Diagnostic Summary Tables ----------------------------

gen_summary_tables <- function(unit_dispatch_sumbase, unit_type, group_vec) {
  
  ## 3A. Summary Table 1 -- total profits [$/kW-mo], operating hours, and starts for each month in each year in each area [not exported]
  g1_vec <- c(group_vec, "Model.Year", "Month")
  month_yr_kwmo <- unit_dispatch_sumbase %>%
    group_by_at(g1_vec) %>%
    summarise(Profit.per.kWmo = sum(RTD.Final.Hourly.Profit, na.rm = TRUE)/(1000),
              Hour.Count = sum(Energy.Flag, na.rm = TRUE),
              Start.Count = sum(Unit.Start, na.rm = TRUE))
  
  ## 3B. Summary Table 2 -- average profits [$/kW-mo], operating hours, and starts for each month in each area. Averages taken from the yearly figures given in Summary Table 1 (3A).
  g2_vec <- c(group_vec, "Month")
  month_kwmo <- month_yr_kwmo %>%
    group_by_at(g2_vec) %>%
    summarise(Avg.Profit.per.kWmo = mean(Profit.per.kWmo),
              Avg.Hour.Count = mean(Hour.Count),
              Avg.Start.Count = mean(Start.Count))
  
  ## 3C. Summary Table 3 -- total profits [$/kW-yr] and total operating hours by DAM commitment and RTD dispatch decision. Averages based on hourly profits.
  
  g3_vec <- c(group_vec, "Model.Year", "DAM.Commit")
  dam_v_rtd_kwyr <- unit_dispatch_sumbase %>%
    group_by_at(g3_vec) %>%
    summarise(Energy.Profit.per.kWyr = sum(Energy.Profit, na.rm = TRUE)/1000,
              Energy.Hours = sum(Energy.Flag, na.rm = TRUE),
              Reserve.Profit.per.kWyr = sum(Reserve.Profit, na.rm = TRUE)/1000,
              Reserve.Hours = sum(Reserve.Flag, na.rm = TRUE),
              Buyout.Profit.per.kWyr = sum(Buyout.Profit, na.rm = TRUE)/1000,
              Buyout.Hours = sum(Buyout.Flag, na.rm = TRUE),
              Limited.Profit.per.kWyr = sum(Limited.Profit, na.rm = TRUE)/1000,
              Limited.Hours = sum(Limited.Flag, na.rm = TRUE),
              None.Profit.per.kWyr = sum(None.Profit, na.rm = TRUE)/1000,
              None.Hours = sum(None.Flag, na.rm = TRUE))  
  
  
  
  ## 3D. Summary Table 4 -- total profits [$/kW-Yr] and total operating hours by RTD dispatch decision.
  g4_vec <- c(group_vec, "Model.Year", "RTD.Dispatch")
  rtd_dispatch_kwyr <- unit_dispatch_sumbase %>%
    group_by_at(g4_vec) %>%
    summarise(Profit.per.kWyr = sum(RTD.Final.Hourly.Profit, na.rm = TRUE)/1000,
              Energy.Hours = sum(Energy.Flag, na.rm = TRUE),
              Reserve.Hours = sum(Reserve.Flag, na.rm = TRUE),
              Buyout.Hours = sum(Buyout.Flag, na.rm = TRUE),
              Limited.Hours = sum(Limited.Flag, na.rm = TRUE),
              None.Hours = sum(None.Flag, na.rm = TRUE))
  
  ## 3D. Summary Table 5 -- total profits [$/kW-Yr] and total operating hours by fuel used (energy dispatch hours only).
  
  g5_vec <- c(group_vec, "Model.Year", "Fuel.Used")
  fuel_type_dispatch_kwyr <- unit_dispatch_sumbase %>%
    group_by_at(g5_vec) %>%
    filter(RTD.Dispatch == "Energy") %>%
    summarise(Energy.Profit.per.kWyr = sum(RTD.Final.Hourly.Profit, na.rm = TRUE)/1000,
              Energy.Hours = sum(Energy.Flag, na.rm = TRUE))  
  
  
  
  if (unit_type != "CC") {
    
    ## 3E. Summary Table 6 -- average profits [$/kW-yr], operating hours, and starts for each area. Averages taken from the monthly averages in Summary Table 2 (3B).
    
    threeyr_kwyr <- month_kwmo %>%
      group_by_at(group_vec) %>%
      summarise(Avg.Profit.per.kWyr = sum(Avg.Profit.per.kWmo),
                Avg.Hour.Count = sum(Avg.Hour.Count),
                Avg.Start.Count = sum(Avg.Start.Count))
    
    
    ## 3F. Summary Table 7 -- total profits [$/kW-yr], operating hours, and starts for each year in each area.
    
    g7_vec <- c(group_vec, "Model.Year")
    yr_kwyr <- dispatch_sumbase %>% 
      group_by_at(g7_vec) %>%
      summarise(Profit.per.kWyr = sum(RTD.Final.Hourly.Profit, na.rm = TRUE)/1000,
                Hour.Count = sum(Energy.Flag, na.rm = TRUE),
                Start.Count = sum(Unit.Start, na.rm = TRUE))
    
  } else if (unit_type == "CC") {
    
    ## 3E. Summary Table 6 -- average profits [$/kW-yr], operating hours, and starts for each area. Averages taken from the monthly averages in Summary Table 2 (3B).
    
    threeyr_kwyr <- month_kwmo %>%
      group_by_at(group_vec) %>%
      summarise(Avg.Profit.per.kWyr = sum(Avg.Profit.per.kWmo),
                Avg.Hour.Count = sum(Avg.Hour.Count),
                Avg.Start.Count = sum(Avg.Start.Count))
    
    ## 3F. Summary Table 7 -- total profits [$/kW-yr], operating hours, and starts for each year in each area.
    
    g7_vec <- c(group_vec, "Model.Year")
    yr_kwyr <- dispatch_sumbase %>% 
      group_by_at(g7_vec) %>%
      summarise(Profit.per.kWyr = sum(RTD.Final.Hourly.Profit)/1000,
                Hour.Count = sum(Energy.Flag, na.rm = TRUE),
                Start.Count = sum(Unit.Start, na.rm = TRUE))
  }
  
  
  results <- list(threeyr_kwyr,
                  yr_kwyr,
                  month_kwmo,
                  dam_v_rtd_kwyr,
                  rtd_dispatch_kwyr,
                  fuel_type_dispatch_kwyr) %>% 
    set_names("3Yr_Avg_$kWYr",
              "Yrly_$kWYr",
              "Mth_Avg_$kWmo",
              "DAMvRTD_$kWYr",
              "RTD_Disp_$kWYr",
              "FuelType_EngOnly_$kWYr")  
  
  
  
  return(results)
}


### 4. Results Table for Export Function ------------
isolate_result_type <- function(case, start_date, result_cat) {
  case_tables <- diagnostic_tables[[case]]
  toggle_list <- case_list[[case]]$Toggles
  all_case_results <- lapply(names(case_tables), 
                             extract_numbers, 
                             tables = case_tables,
                             result_cat = result_cat,
                             case_name = toggle_list$Case_ID,
                             SCR = toggle_list$SCR_Toggle,
                             Gas_Only = toggle_list$Gas_Only_Toggle,
                             Zone_C_Hub = toggle_list$Zone_C_Hub,
                             Price_Cost_Inputs = toggle_list$Price_Cost_Inputs_Filename,
                             Technical_Inputs = toggle_list$Technical_Inputs_Filename,
                             start_date = start_date) %>% 
    rbindlist()
  return(all_case_results)
}
extract_numbers <- function(unit, tables, result_cat, case_name, SCR, Gas_Only, Zone_C_Hub, Price_Cost_Inputs, Technical_Inputs, start_date){
  results <- tables[[unit]][[result_cat]] %>%
    mutate(Unit = unit, Case = case_name, Start_Date = as.Date(start_date), 
           SCR_Toggle = SCR, Gas_Only_Toggle = Gas_Only,
           Zone_C_Hub = Zone_C_Hub,
           Price_Cost_Inputs = Price_Cost_Inputs,
           Technical_Inputs = Technical_Inputs)
  results <- select(results, Case, Unit, Area, everything())
  return(results)
} 

